//------------------------------------------------------
// module  : GaitAnalysis.cc
// author  : Mignotte Max
// date    : 
// version : 1.0
// language: C++
// note    :
//------------------------------------------------------
//  

//------------------------------------------------
// INCLUDED FILES --------------------------------
//------------------------------------------------
#include "FunctionUseful.h"
#include "Chrono.h"

#include "PreProcDepthMapSeq.h"
#include "MdsBasedSymEst.h"

//------------------------------------------------
// FOR WINDOWS -----------------------------------
//------------------------------------------------
#include <X11/Xutil.h>
#include <X11/Xos.h>

Display   *display;
int	  screen_num;
int 	  depth;
Window	  root;
Visual*	  visual;
GC	  gc;

//------------------------------------------------
// CONSTANTES ------------------------------------
//------------------------------------------------
const int NBVISUSEQ=1;

const int NBFRAMES=300;
const int OFFSET=500;
const int FLAGMEDIAN=0;

const int REFINING=1;
const int DISTANCE=1;

const int ZOOM=1;
const int QUIT=0;
const int ASICURVE=0;

//------------------------------------------------
// VARIABLES GLOBALES ----------------------------
//------------------------------------------------
char Name_Img[100];
char Extension[100];

int NbFrames;
int Offset;
int FlagMedian;

int Refining;
int Stochast;
int Distance;
float FactPrior;

int zoom;
int flag_quit;
int flag_AsiCurve;

//-----------------------------------------------------------------------
// OPEN_DISPLAY()							
//-----------------------------------------------------------------------
int open_display()
{
  if ((display=XOpenDisplay(NULL))==NULL)
   { printf("Connection impossible\n");
     return(-1); }

  else
   { screen_num=DefaultScreen(display);
     visual=DefaultVisual(display,screen_num);
     depth=DefaultDepth(display,screen_num);
     root=RootWindow(display,screen_num);
     return 0; }
}

//-----------------------------------------------------------------------
// FABRIQUE_WINDOW()							
// Cette fonction cre une fenetre X et l'affiche  l'cran.	        
//-----------------------------------------------------------------------
Window fabrique_window(char *nom_fen,int x,int y,int width,int height,int zoom)
{
  Window                 win;
  XSizeHints      size_hints;
  XWMHints          wm_hints;
  XClassHint     class_hints;
  XTextProperty  windowName, iconName;

  char *name=nom_fen;

  if(zoom<0) { width/=-zoom; height/=-zoom; }
  if(zoom>0) { width*=zoom;  height*=zoom;  }

  win=XCreateSimpleWindow(display,root,x,y,width,height,1,0,255);

  size_hints.flags=PPosition|PSize|PMinSize;
  size_hints.min_width=width;
  size_hints.min_height=height;

  XStringListToTextProperty(&name,1,&windowName);
  XStringListToTextProperty(&name,1,&iconName);
  wm_hints.initial_state=NormalState;
  wm_hints.input=True;
  wm_hints.flags=StateHint|InputHint;
  class_hints.res_name=nom_fen;
  class_hints.res_class=nom_fen;

  XSetWMProperties(display,win,&windowName,&iconName,
                   NULL,0,&size_hints,&wm_hints,&class_hints);

  gc=XCreateGC(display,win,0,NULL);

  XSelectInput(display,win,ExposureMask|KeyPressMask|ButtonPressMask| 
               ButtonReleaseMask|ButtonMotionMask|PointerMotionHintMask| 
               StructureNotifyMask);

  XMapWindow(display,win);
  return(win);
}

//--------------------------------------------------------------------------
// CREE_XIMAGENB()							    
// Cre une XImage  partir d'un tableau de float                           
// L'image peut subir un zoom.						    
//--------------------------------------------------------------------------
XImage* cree_XimageNB(unsigned short int** mat,int z,int length,int width)
{
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  unsigned char	 pix;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positiv*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pix=(unsigned char)mat[lig/z][col/z];
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pix;
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=pix; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatifv*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    somme=0.0;
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     somme+=mat[lig+zoom_lig][col+zoom_col];
           
     somme/=(z*z);    
     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)somme;
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)somme; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

//---------------------------------------------------------------------------
// CREE_XIMAGECOUL()							    
// Cre une XImage  partir d'un tableau 3D de  unsigned short int                        
// L'image peut subir un zoom.						    
//--------------------------------------------------------------------------
XImage* cree_XimageCoul(unsigned short int*** matRVB,int z,int length,int width)
{
  int i;
  int lgth,wdth,lig,col,zoom_col,zoom_lig;
  float somme;
  float sum[3];
  unsigned char	 pixR,pixV,pixB,pixN;
  unsigned char* dat;
  XImage* imageX;

  /*Zoom positif*/
  /*------------*/
  if (z>0)
  {
   lgth=length*z;
   wdth=width*z;

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<lgth;lig=lig+z) for(col=0;col<wdth;col=col+z)
   { 
    pixR=(unsigned char)matRVB[0][lig/z][col/z];
    pixV=(unsigned char)matRVB[1][lig/z][col/z];
    pixB=(unsigned char)matRVB[2][lig/z][col/z];
    somme=(1.0/3.0)*(pixR+pixV+pixB);
    pixN=(unsigned char)somme;

    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
      { 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+0)]=pixB; 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+1)]=pixV; 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+2)]=pixR; 
       dat[((lig+zoom_lig)*wdth*4)+((4*(col+zoom_col))+3)]=0; 
       }
    }
  } /*--------------------------------------------------------*/

  /*Zoom negatif*/
  /*------------*/
  else
  {
   z=-z;
   lgth=(length/z);
   wdth=(width/z);

   dat=(unsigned char*)malloc(lgth*(wdth*4)*sizeof(unsigned char));
   if (dat==NULL)
      { printf("Impossible d'allouer de la memoire.");
        exit(-1); }

  for(lig=0;lig<(lgth*z);lig=lig+z) for(col=0;col<(wdth*z);col=col+z)
   {  
    sum[0]=sum[1]=sum[2]=0.0;
    
    for(i=0;i<3;i++)
    for(zoom_lig=0;zoom_lig<z;zoom_lig++) for(zoom_col=0;zoom_col<z;zoom_col++)
     sum[i]+=matRVB[i][lig+zoom_lig][col+zoom_col];
       
    for(i=0;i<3;i++)  sum[i]/=(z*z); 

     dat[((lig/z)*wdth*4)+((4*(col/z))+0)]=(unsigned char)sum[1];
     dat[((lig/z)*wdth*4)+((4*(col/z))+1)]=(unsigned char)sum[1];
     dat[((lig/z)*wdth*4)+((4*(col/z))+2)]=(unsigned char)sum[1];
     dat[((lig/z)*wdth*4)+((4*(col/z))+3)]=(unsigned char)sum[1]; 
   }
  } /*--------------------------------------------------------*/

  imageX=XCreateImage(display,visual,depth,ZPixmap,0,(char*)dat,wdth,lgth,16,wdth*4);
  return (imageX);
}

//---------------------------------------------------------
// Enregistre les arguments de la ligne de commande      
//---------------------------------------------------------
void read_arguments(int argc,char** argv)
{
 int i;

   //Options
  if (argc<2) 
    { printf("\n Usage %s [Img. in ppm format]",argv[0]); 
      printf("\n Options : value by default indicated in []");
      printf("\n            -----------------------------------------");
      printf("\n            -------------");
      printf("\n            -f NbFrames          >[%d]",NBFRAMES);
      printf("\n            -o Offset            >[%d]",OFFSET);
      printf("\n            -------------");
      printf("\n            -d Dist.             >[%d]",DISTANCE);
      printf("\n               [L2NS:0] [L2:1] [L1:2] [LInf:3] [LMod:4] [LRad.:5] [LGrd=6] [LMoy=7] > [:%d])",DISTANCE);
      //printf("\n            -p Fact.Prior        >[%.3f]",FACT_PRIOR); 
      printf("\n            -------------");
      printf("\n            -m FlagMedian -Avec- >[%d]",FLAGMEDIAN); 
      printf("\n            -s Stochastic -Avec- >[%d]",STOCHAST); 
      printf("\n            -r Refining   -Sans- >[%d]",REFINING); 
      printf("\n            -------------");
      printf("\n            -g <Extension>  -Sans-");
      printf("\n            -------------");
      printf("\n            -z zoom (%d)",ZOOM);
      printf("\n            -q quit (Sans)"); 
      printf("\n            -a avec courbe ASI");
      printf("\n            -------------");
      printf("\n");
      printf("\n  Example: %s DATA/S05_A.dat -f 300 -o 400",argv[0]);
      printf("\n\n\n\n\n");
      exit(-1); }

      //Chargement Fichier
      if (argc>1) strcpy(Name_Img,argv[1]);
     

   //Boucle 
   for(i=2;i<argc;i++)
     {   
          switch(argv[i][1])
          {  
           case 'f': NbFrames=atoi(argv[++i]);     break;
           case 'o': Offset=atoi(argv[++i]);       break;
	   
           case 'd': Distance=atoi(argv[++i]);     break;
	   case 'p': FactPrior=atof(argv[++i]);    break;

           case 'm': FlagMedian=1;                 break;
           case 'r': Refining=0;                   break;
	   case 's': Stochast=1;                   break;

           case 'g': strcpy(Extension,argv[++i]);  break;
                    
	   case 'z': zoom=atoi(argv[++i]);         break;
	   case 'q': flag_quit=1;                  break;     
           case 'a': flag_AsiCurve=1;              break;             
          }    
     } 
 }

//----------------------------------------------------------
//----------------------------------------------------------
// PROGRAMME PRINCIPAL -------------------------------------
//----------------------------------------------------------
//----------------------------------------------------------
int main(int argc,char** argv)
{
 int k;
 int flag;
 float corr;

 //>Pour Xwindow
 Window win_ppicture_SeqWalk=0;
 XImage *x_ppicture_SeqWalk;
 XImage *x_ppicture_SymMap;
 char nomfen_ppicture_SeqWalk[100];
 
 //>Initialisation
 NbFrames=NBFRAMES;
 Offset=OFFSET;

 Distance=DISTANCE;
 FactPrior=FACT_PRIOR;

 FlagMedian=FLAGMEDIAN;
 Refining=REFINING;
 Stochast=STOCHAST;
 
 zoom=ZOOM;
 flag_quit=QUIT;
 flag=0;
 flag_AsiCurve=ASICURVE;

 strcpy(Extension,"");

 //>Read Arguments
 read_arguments(argc,argv);

 //>Display Options
 printf("\n Information: Image[%s]",Name_Img); 
 printf("\n -----------");  
 printf("\n NbFrames   : [%d]",NbFrames);
 printf("\n Offset     : [%d]",Offset);
 printf("\n -----------");
 printf("\n Distance : [%d]",Distance);
 printf("\n FactPrior : [%.4f]",FactPrior);
 printf("\n -----------");
 printf("\n FlagMedian : [%d]",FlagMedian);
 printf("\n Refining : [%d]",Refining);
 printf("\n Stochast : [%d]",Stochast);
 printf("\n -----------");
 printf("\n Extension : [%s]",Extension);
 printf("\n -----------");
 printf("\n Zoom: %d",zoom);
 if (flag_quit)      printf("\n -> Avec Quit");
 if (flag_AsiCurve)  printf("\n -> Avec Courbe ASI");
 printf("\n");
 fflush(stdout);

 //>Chrono
 Chrono chrono;
 chrono.reset();
 int timesec;

 //------
 //------------------------------------------
 //->-[0] Memory Allocation 
 //------------------------------------------
 //>Memory Allocation
 unsigned short int*** RawDepthMapImgSeq=usimatrix_allocate_3d(NbFrames,HEIGHT,WIDTH);
 unsigned short int*** ProcDepthMapImgSeq=usimatrix_allocate_3d(NbFrames,HEIGHT,WIDTH);
 unsigned short int*** RawMdsSymMap=usimatrix_allocate_3d(TROIS,HEIGHT,WIDTH);
 unsigned short int*** ProcMdsSymMap=usimatrix_allocate_3d(TROIS,HEIGHT,WIDTH);
 unsigned short int*** RGBMdsSymMap=usimatrix_allocate_3d(TROIS,HEIGHT,WIDTH);

 //------
 //------------------------------------------
 //->-[X] Tests
 //------------------------------------------
 //Test-1
 //printf("\n >-[X] TEST-1"); 
 //MdsBasedSymEst SymEstTest(Name_Img,ProcDepthMapImgSeq,HEIGHT,WIDTH,NbFrames);
 //SymEstTest.VisuThreeClusters();
 //exit(-1);

 //Test-2
 //printf("\n >-[X] TEST-2");
 //MdsBasedSymEst SymEstTest(Name_Img,ProcDepthMapImgSeq,HEIGHT,WIDTH,NbFrames);
 //SymEstTest.DistMap=Distance;
 //SymEstTest.ReduceCurveASI_2Dpts();
 //exit(-1);

 //------
 //------------------------------------------
 //->-[1] Read DepthMapSequence
 //------------------------------------------
 //>File Path_&_Name
 char ImgPathAndName[NCHAR];
 strcpy(ImgPathAndName,PATH_DEPTHMAPSEQ);
 strcat(ImgPathAndName,Name_Img);
 printf("\n >-[1] Read DepthMapSequence Name File: %s\n",ImgPathAndName); 

 //>Read File
 ReadRawDepthMapSeqFile(ImgPathAndName,RawDepthMapImgSeq,NbFrames,Offset);
 CopyMatrix(RawDepthMapImgSeq,ProcDepthMapImgSeq,NbFrames,HEIGHT,WIDTH);
 

 //------
 //------------------------------------------
 //->-[2] PreProcess DepthMapSequence
 //------------------------------------------
 //>Obj.Pre-Processing
 printf("\n >-[2] Pre-Processing File Name [%s][%d:%d:%d]",Name_Img,NbFrames,HEIGHT,WIDTH);
 PreProcDepthMapSeq preprocess(Name_Img,HEIGHT,WIDTH,NbFrames);

 //>Silouhette Extraction
 preprocess.FlagMedian=FlagMedian;
 preprocess.Segmente_Silouhette(RawDepthMapImgSeq,ProcDepthMapImgSeq,NORM_FACTOR,MEDIAN_W_SIZE);
 
 //>VisuXWindows [ImgSeq]
 if (!flag_quit)
 { //>Start Graphical Session 
   if (open_display()<0) printf(" Impossible d'ouvrir une session graphique");
   sprintf(nomfen_ppicture_SeqWalk,"Image : %s",Name_Img);
   win_ppicture_SeqWalk=fabrique_window(nomfen_ppicture_SeqWalk,10,10,WIDTH,HEIGHT,zoom);

   //>LoopOnSequence
   for(k=0;k<(NBVISUSEQ*NbFrames);k++) 
         { //if (!(k%NbFrames)) printf("[%d]",k/NbFrames);
           fflush(stdout);
           x_ppicture_SeqWalk=cree_XimageNB(ProcDepthMapImgSeq[k%(NbFrames)],zoom,HEIGHT,WIDTH);
           XPutImage(display,win_ppicture_SeqWalk,gc,x_ppicture_SeqWalk,0,0,0,0,
                     x_ppicture_SeqWalk->width,x_ppicture_SeqWalk->height);
           usleep(10000); 
           XDestroyImage(x_ppicture_SeqWalk); } } 
   

 //------  
 //--------------------------------------------
 //->-[3] Algo>MDS-Based Symetry Estimation
 //--------------------------------------------
 //Presentation
 printf("\n >-[3] MDS-based Symmetry Estimation [File: %s]",ImgPathAndName); 
 fflush(stdout);

 //>Obj.MdsBasedSymEst_&_Options
 MdsBasedSymEst SymEst(Name_Img,ProcDepthMapImgSeq,HEIGHT,WIDTH,NbFrames);
 SymEst.FlgRefining=Refining;
 SymEst.FactPrior=FactPrior;
 SymEst.FlgStoch=Stochast;
 strcpy(SymEst.Extension,Extension);
 strcpy(SymEst.ExtensionASI,Extension);

 //>MDS-RoughEstimation
 SymEst.MdsRoughEstimation(Distance,RawMdsSymMap);

 //>VisuXWindows_[Raw Sym. Map]
 if (!flag_quit)
    { x_ppicture_SymMap=cree_XimageCoul(RawMdsSymMap,zoom,HEIGHT,WIDTH);
      XPutImage(display,win_ppicture_SeqWalk,gc,x_ppicture_SymMap,0,0,0,0,
		x_ppicture_SymMap->width,x_ppicture_SymMap->height); }

 //>Corr.Stats
 printf("\n > Correlation Before Refining:"); 
 fflush(stdout);
 corr=CompCorrMetric(ProcDepthMapImgSeq,RawMdsSymMap,NbFrames,2,2,\
                     PATIENT_XMIN,PATIENT_XMAX,PATIENT_YMIN,PATIENT_YMAX); 
 printf("[corr=%.3f]\n",corr);
 SymEst.CorelRaw=corr;
 timesec=chrono.elapsed(SECOND); chrono.print();

 //>Save
 //SaveImagePpm("",(char*)"RawMdsSymMap",(char*)"",RawMdsSymMap,HEIGHT,WIDTH);


 //------
 //--------------------------------------------
 //->- [4] Refining Map
 //--------------------------------------------
 if (Refining)
    {
     //>ComputeNonLocGraph
     printf("\n >-[4a] Refining Map: Comp. NonLocGraph [Cnx:%d][WdSz:%d] (Wait)... ",NBCNXGRAPH,WINDSIZEEXPLOR);
     fflush(stdout);
     SymEst.ComputeNonLocalGraph(NBCNXGRAPH,WINDSIZEEXPLOR);
 
     //>Minimization_LocalRefining
     SymEst.MetropolisRefine(ProcMdsSymMap,TEMPINIT,TEMPFINA,RADIUS,MAX_ITER);

     //>VisuXWindows_[ProcMdsSymMap]
     if (!flag_quit)
        { x_ppicture_SymMap=cree_XimageCoul(ProcMdsSymMap,zoom,HEIGHT,WIDTH);
          XPutImage(display,win_ppicture_SeqWalk,gc,x_ppicture_SymMap,0,0,0,0,
		    x_ppicture_SymMap->width,x_ppicture_SymMap->height); }
 
     //>Corr.Stats
     printf(" > Correlation After Refining:"); 
     fflush(stdout);
     corr=CompCorrMetric(ProcDepthMapImgSeq,ProcMdsSymMap,NbFrames,2,2,\
                         PATIENT_XMIN,PATIENT_XMAX,PATIENT_YMIN,PATIENT_YMAX); 
     printf("[corr=%.3f]\n",corr);
     SymEst.CorelRefined=corr;

     //>Save
     strcat(SymEst.Info,"_Refined");
     //SaveImagePpm("",(char*)"ProcMdsSymMap",SymEst.Info,ProcMdsSymMap,HEIGHT,WIDTH);
    }

 //>LAB>RGB(_non_saturated)
 printf("\n >-[4b] Conversion LAB_TO_RGB : LAB>");
 SymEst.ColorConvLABtoRGBNonSat(RGBMdsSymMap);

 //>VisuXWindows_[RGB Proc.Sym.Map]
 if (!flag_quit)
    { x_ppicture_SymMap=cree_XimageCoul(RGBMdsSymMap,zoom,HEIGHT,WIDTH);
      XPutImage(display,win_ppicture_SeqWalk,gc,x_ppicture_SymMap,0,0,0,0,
		x_ppicture_SymMap->width,x_ppicture_SymMap->height); }

 

 //------
 //--------------------------------------------
 //->- [5] Compute Index
 //--------------------------------------------
 //>Compute_ASI
 float ASI_idx=SymEst.EstimateASI(SymEst.ProcMdsSymMap);
 printf("<%.2f>\n",ASI_idx);
 fflush(stdout);

 //>ElapsedTime
 printf("\n"); 
 timesec=chrono.elapsed(SECOND); chrono.print();
 SymEst.ElapsedTime=timesec;
 
 //>FinalResumeInfo
 SymEst.GiveResumeInfo();

 //>Save
  SymEst.SaveImagePpm("",(char*)"RGBMdsSymMap",SymEst.Info,RGBMdsSymMap,HEIGHT,WIDTH);
 if (flag_AsiCurve) SymEst.SaveCurveASI();


 //------
 //-------------
 //End
 //-------------
 //FreeMemory
 if (RawDepthMapImgSeq)  free_usimatrix_3d(RawDepthMapImgSeq,NbFrames); 
 if (ProcDepthMapImgSeq) free_usimatrix_3d(ProcDepthMapImgSeq,NbFrames); 
 if (RawMdsSymMap)       free_usimatrix_3d(RawMdsSymMap,TROIS); 
 if (ProcMdsSymMap)      free_usimatrix_3d(ProcMdsSymMap,TROIS); 
 if (RGBMdsSymMap)       free_usimatrix_3d(RGBMdsSymMap,TROIS); 

 //ReturnWithoutProblem
 printf("\n C'est fini... \n"); 
 fflush(stdout);
 if (!flag_quit) for(;;) { ; }
 return 0;
 }
 


