//---------------------------------------------------
// module  : f_FastMap.cc                 
// auteur  : Max Mignotte, Antoine Moevus /Christos                      
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------
// From:
/* $Revision: 7.3 $ */
/* $Author: christos $ */
/* $Id: vector.h,v 7.3 1995/03/17 00:20:17 christos Exp christos $ */

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "FastMap.h"

//----------------------------------------------------------
// Constructor/Destructor FastMap class
//----------------------------------------------------------
FastMap::FastMap() { ; }
FastMap::~FastMap() { ; }

//------------------------------------------------
// FastMap class METHOD
//------------------------------------------------
VECTOR *FastMap::vecalloc( int n){
    VECTOR *pvec;

    assert( n > 0 );
    pvec = (VECTOR *) malloc( sizeof(VECTOR) );
    assert( pvec != NULL );
    pvec-> vec = (NUMBER *) malloc ( sizeof(NUMBER) * n );
    assert( (pvec->vec) != NULL );
    pvec->len = n;

    return(pvec);
}

void FastMap::vecfree( VECTOR *pvec){
    free( (char *) pvec->vec);
    pvec->len =0;
    free( (char *) pvec);
}

NUMBER FastMap::vecdist2( VECTOR *p1, VECTOR *p2){
    NUMBER res2; /* squared result */
    NUMBER v1, v2;
    int pos, length;

    assert( (p1->len) == (p2->len) );
    length = p1->len;
    res2=0.0;
    for(pos=0; pos< length; pos++){
        v1 = (p1->vec)[pos];
        v2 = (p2->vec)[pos];
        res2 += (v1-v2)*(v1-v2);
    }

    return( res2);
}

void FastMap::vecput( VECTOR *pvec, int pos, NUMBER val){
    assert( pos >=0) ;
    assert( pos < (pvec->len)  );
    (pvec->vec)[pos] =  val;
}

NUMBER FastMap::vecget( VECTOR *pvec, int pos){
    assert( pos >=0) ;
    assert( pos < (pvec->len)  );
    return( (pvec->vec)[pos] );

}

VECTOR *FastMap::veccopy( VECTOR *psource){
    VECTOR *ptarget;
    int pos, length;

    assert( (psource->len) >0 );
    length = psource->len;
    ptarget = vecalloc(length);
    /* printf( "%d\n" , ptarget->len); */
    for(pos=0; pos<length; pos++){
        (ptarget->vec)[pos] = (psource->vec)[pos];
    }
    return( ptarget);
}

void FastMap::vecprint( VECTOR *p){
    int i, length;

    /*  assert( (p->len) > 0 ); */
    length = p->len;

    for(i=0; i< length; i++){
        printf("%g\t", vecget(p, i));
    }
    printf("\n");
}
/*===================================================*/


/*===================================================*/
OBJECT *FastMap::oballoc( int n){
    OBJECT *pob;

    assert( n > 0 );
    pob = (OBJECT *) malloc( sizeof(OBJECT) );
    assert( pob != NULL );
    pob-> vec = (NUMBER *) malloc ( n * sizeof(NUMBER) );
    assert( (pob->vec) != NULL );
    pob->len = n;

    return(pob);
}

void FastMap::obfree( OBJECT *pob){
    free( (char *) pob->vec);
    pob->len =0;
    free( (char *) pob);
}



/*===================================================*/
// Original euclidean distance
/*===================================================*/

//NUMBER obdist2( OBJECT *p1, OBJECT *p2){
//    NUMBER res2; // squared result
//    NUMBER v1, v2;
//    int pos, length;

//    assert( (p1->len) == (p2->len) );
//    length = p1->len;

//    res2=0.0;
//    for(pos=0; pos< length; pos++){
//        v1 = (p1->vec)[pos];
//        v2 = (p2->vec)[pos];
//        res2 += (v1-v2)*(v1-v2);
//    }

//    return(sqrt(res2));
//}



void FastMap::obput( OBJECT *pob, int pos, NUMBER val){
    assert( pos >=0) ;
    assert( pos < (pob->len)  );
    (pob->vec)[pos] =  val;
}

NUMBER FastMap::obget( OBJECT *pob, int pos){
    assert( pos >=0) ;
    assert( pos < (pob->len)  );
    return( (pob->vec)[pos] );

}

OBJECT *FastMap::obcopy( OBJECT *psource){
    OBJECT *ptarget;
    int pos, length;

    assert( (psource->len) >0 );
    length = psource->len;
    ptarget = oballoc(length);
    for(pos=0; pos<length; pos++){
        (ptarget->vec)[pos] = (psource->vec)[pos];
    }
    return( ptarget);
}

void FastMap::obprint( OBJECT *p){
    int i, length;

    /*  assert( (p->len) > 0 ); */
    length = p->len;

    for(i=0; i< length; i++){
        printf("%g\t", obget(p, i));
    }
    printf("\n");
}

// OBJECT *FastMap::line2ob(char *line){
//     char *tmpline;  /* temp copy */
//     OBJECT *pob;
//     char *token;
//     int ntokens;
//     NUMBER val;
//     int i;

//     ntokens = counttokens(line);
//     pob = oballoc(ntokens);

//     /* copy, because strtok() messed up tmpline */
//     tmpline=strdup(line);

//     assert(ntokens>0); /* no empty lines */
//     i=0;
//     token = strtok(tmpline, BLANKS);
//     /* atof did not work! */
//     /* val = (NUMBER) atof( token); */
//     sscanf( token, "%lf", &val);
//     obput(pob, i, val);
//     for(i=1; i< ntokens; i++){
//         token = strtok( (char *)NULL, BLANKS);
//         sscanf( token, "%lf", &val);
//         obput(pob, i,val);
//     }
//     free(tmpline);
//     return(pob);
// }
/*===================================================*/


/*===================================================*/
MATRIX *FastMap::matrixalloc( int n){
    MATRIX *pmatrix;

    assert( n > 0 );
    pmatrix = (MATRIX *) malloc( sizeof(MATRIX) );
    assert( pmatrix != NULL );
    pmatrix-> list = (VECTOR **) malloc ( sizeof(VECTOR *) * n );
    assert( (pmatrix->list) != NULL );

    pmatrix->count = n;

    /* ******** DO NOT allocate space for vec ****
    for(i=0; i<count; i++){
        (pmatrix->list)[i] = vecalloc(m);
    }
    ******** */

    return(pmatrix);
}

void FastMap::matrixfree( MATRIX *pmatrix){
    int i;
    for(i=0; i< (pmatrix->count); i++){
        vecfree( (pmatrix->list)[i] );
    }
    free( (char *) pmatrix->list);
    pmatrix->count =0;
    free( (char *) pmatrix);
}

void FastMap::matrixput( MATRIX *pmatrix, int pos, VECTOR *pvec){
    assert( pos >=0) ;
    assert( pos < (pmatrix->count)  );
    (pmatrix->list)[pos] = veccopy(pvec);
}

void FastMap::matrix2put( MATRIX *pmatrix, int row, int column, NUMBER val){
    VECTOR *pvec;
    assert( row >=0);
    assert( row < (pmatrix->count)  );
    pvec = (pmatrix->list)[row];
    vecput( pvec, column, val);
}

VECTOR *FastMap::matrixget( MATRIX *pmatrix, int pos){
    assert( pos >=0) ;
    assert( pos < (pmatrix->count)  );
    return( (pmatrix->list)[pos] );
}

NUMBER FastMap::matrix2get( MATRIX *pmatrix, int row, int column){
    VECTOR *pvec;
    NUMBER val;
    assert( row >=0);
    assert( row < (pmatrix->count)  );
    pvec = (pmatrix->list)[row];
    val = vecget( pvec, column);
    return(val);
}

void FastMap::matrixprint( MATRIX *pmatrix){
    int i, n;
    if( pmatrix == NULL ){
        printf("empty matrix\n");
        return;
    }
    assert( pmatrix->count >0 );
    n = pmatrix->count;
    for(i=0; i<n; i++){
        vecprint( matrixget( pmatrix, i) );
    }
}

/* prints pairwise squared distances */
void FastMap::printvdist2(MATRIX *pmat){
    int i, j, n;

    n = pmat->count;
    for(i=0; i<n; i++){
        for(j=0; j<n; j++){
            printf("%g\t", vecdist2( matrixget(pmat, i),
                                     matrixget(pmat, j) ) );
        }
        printf("\n");
    }
}
/*===================================================*/


/*===================================================*/
OA *FastMap::oaalloc( int n){
    OA *poa;

    assert( n > 0 );
    poa = (OA *) malloc( sizeof(OA) );
    assert( poa != NULL );
    poa-> list = (OBJECT **) malloc ( sizeof(OBJECT *) * n );
    assert( (poa->list) != NULL );

    poa->count = n;

    /* ******** DO NOT allocate space for objects ****
    for(i=0; i<count; i++){
        (poa->list)[i] = oballoc(m);
    }
    ******** */

    return(poa);
}

void FastMap::oafree( OA *poa){
    int i;
    for(i=0; i< (poa->count); i++){
        obfree( (poa->list)[i] );
    }
    free( (char *) poa->list);
    poa->count =0;
    free( (char *) poa);
}

/* copies an object to the given position */
void FastMap::oaput( OA *poa, int pos, OBJECT *pob){
    assert( pos >=0) ;
    assert( pos < (poa->count)  );
    (poa->list)[pos] = obcopy(pob);
}

OBJECT *FastMap::oaget( OA *poa, int pos){
    assert( pos >=0) ;
    assert( pos < (poa->count)  );
    return( (poa->list)[pos] );
}

void FastMap::oaprint( OA *poa){
    int i, n;
    assert( poa->count > 0 );
    n = poa->count;
    for(i=0; i<n; i++){
        obprint( oaget( poa, i ) );
    }
}

/*===================================================*/


/*===================================================*/
void FastMap::myrewind(FILE *fp){
    if( ferror(fp) ){ clearerr(fp);}
    rewind(fp);
    return;
}

int FastMap::countlines(FILE *fp){
    int nlcount = 0;
    int ichar;

    myrewind(fp);
    while( (ichar=fgetc(fp)) != EOF ){
        if( ichar == '\n' ){ nlcount ++ ;}
    }
    myrewind(fp);

    return(nlcount);
}

/* measures the number of columns of the first line */
int FastMap::countcolumns(FILE *fp){
    char *line;

    myrewind(fp);

    line= getline(fp);
    return( counttokens( line) );
}

void FastMap::checklength(char *line){
    if( index(line, '\n') == NULL ){ /* too long */
        fprintf(stderr, "too long line - stopping \n");
        exit(-1);
    }
}

int FastMap::counttokens(char *line){
    char *templine;
    int res = 0 ;

    /* create a copy, because strtok() messes its arg */
    templine = strdup(line);

    if( strtok(templine, BLANKS) != NULL){
        res ++ ;
    }
    while( strtok( (char *) NULL, BLANKS ) != NULL){
        res ++ ;
    }
    free( templine);
    return(res);
}

/* reads a line from fp, and checks its length */
char *FastMap::getline(FILE *fp){
    static char line[MAXLINE];

    if( fgets(line, MAXLINE, fp) != '\0'){ /* EOF */
        checklength(line);
    }
    return(line);
}

/* checks to see if all the lines of the file have the same
   number of tokens */

BOOLEAN FastMap::checkfile(FILE *fp){
    char *line;
    int nold, nnew;
    int linecount;
    BOOLEAN kosher = TRUE;

    myrewind(fp);

    /* read in the first line */
    line= getline(fp);
    linecount =1;
    if( feof(fp) ){
        return(-1); /*empty file */
    }else{
        nold = counttokens(line);
#ifdef DEBUG
        /* printf("%d:\t%s", nold, line); */
#endif
    }

    line = getline(fp);
    while( !feof(fp) ){
        linecount ++;
        nnew = counttokens(line);
#ifdef DEBUG
        /* printf("%d:\t%s", nnew, line); */
#endif
        if( nnew != nold){
            kosher = FALSE;
            fprintf(stderr, "line %d has %d tokens; line %d has %d\n",
                    1, nold, linecount, nnew);
        }
        /* assert( nnew == nold ); */
        line = getline(fp);
    }

    myrewind(fp);
    return (kosher);
}

/*===================================================*/




/*===================================================*/
/********************************************
   given an array of objects `*poa'
   and a desirable dimensionality `k'
   it returns an n*k matrix with coordinates
   of points, whose pairwise distances approximate
   the distances of the objects of `*poa'
 ***********************************************/

MATRIX *FastMap::fastmap( OA *poa, int k){
    int n = poa->count;
    int i, a1, a2;
    int dims;
    NUMBER d12sq, d1xsq, d2xsq, x;
    MATRIX *pmat;
    VECTOR *pvec;

    pmat = matrixalloc(n);

    /* blank out *pmat */
    pvec = vecalloc( k );
    for( i=0; i<k; i++){
        /* blank out vector */
        vecput( pvec, i, 0.0);
    }
    for(i=0; i<n; i++){
        /* fill up matrix */
        matrixput( pmat, i, pvec);
    }
    vecfree(pvec);
    /* SHOULD MAKE THE ABOVE INTO  A FUNCTION */

    for( dims=0; dims<k; dims++){

        /* pick two remote objects */
        pickAnchors( poa, &a1, &a2, dims, pmat);

        /* compute projections of each object
      on the axis defined by the two anchors */

        d12sq = mydist2( poa, a1, a2, dims, pmat);
        if( d12sq == 0.0 ) { return(pmat); }
        /* we've exceeded the dimensionality of the data space */

        for( i=0; i<n; i++){
            d1xsq = mydist2( poa, i, a1, dims, pmat);
            d2xsq = mydist2( poa, i, a2, dims, pmat);
            x = (d1xsq + d12sq - d2xsq) / ( 2.0 * sqrt( (double) d12sq ) );
            matrix2put( pmat, i, dims, x);
        } /* end for i */
    } /* end for dims */

    return( pmat );
}


/* choose to far-away objects */
void FastMap::pickAnchors( OA * poa, int *pa1, int *pa2, int dims, MATRIX *pmat){
    int i,j;

    i=0;
    j = Furthest( poa, i, dims, pmat);
    i = Furthest( poa, j, dims, pmat);

    *pa1 = i;
    *pa2 = j;
    return;
}

/* for a given object 'i', returns the index of the furthest one */
int FastMap::Furthest( OA *poa, int icurrent, int dims, MATRIX *pmat){
    NUMBER max = 0.0;
    int furthest= icurrent;
    int n = poa->count;
    int j;

    for(j=0; j<n; j++){
        if (mydist2( poa, icurrent, j, dims, pmat) > max){
            max = mydist2( poa, icurrent, j, dims, pmat);
            furthest = j;
        }
    }
    return(furthest);
}

///* squared distance of two objects, after dims-1 projections */
NUMBER FastMap::mydist2( OA *poa, int o1, int o2, int dims, MATRIX *pmat){
    OBJECT *p1 = oaget( poa, o1);
    OBJECT *p2 = oaget( poa, o2);
    NUMBER res2 = obdist2( p1, p2);

    /* decrease it, by the projections */
    int i;
    for(i=0; i< dims; i++){
        NUMBER temp = matrix2get(pmat, o1, i) - matrix2get(pmat, o2, i);
        res2 -=  temp*temp;
    }
    return(res2);
}

/* prints the  squared distances of all pairs of elements */
void FastMap::printdist2(OA *poa){
    int i, j, n;
    n = poa->count;
    for(i=0; i< n; i++){
        for(j=0; j<n; j++){
            printf("%g\t", obdist2( oaget(poa, i), oaget(poa, j) ));
        }
        printf("\n");
    }
}

/* reads in a file, into an object array */
// OA *FastMap::file2oa(FILE *fp){
//     OA *poa;
//     OBJECT *pob;
//     int i, n;
//     char *line;

//     myrewind(fp);
//     n = countlines(fp);

//     assert( n>0); /* no empty file */

//     poa = oaalloc(n);
//     assert( poa != NULL);

//     myrewind(fp);
//     i=0;
//     line = getline(fp);
//     while( !feof(fp) ){
//         pob = line2ob(line);
//         oaput( poa, i, pob);
//         obfree(pob);

//         i++;
//         line = getline(fp);
//     }

//     assert( i == n);
//     for(i=0; i<n; i++){
//         /*check the lengths of the objects for equality */
//     }

//     return(poa);
// }
/*===================================================*/



/*===================================================*/
/*   MODIFICATION::: M. Mignotte::: 18 August 2010   */
/*===================================================*/

//Pour Distance de Bhattacharyaa

// squared distance of two objects, after dims-1 projections 
//NUMBER mydist2( OA *poa, int o1, int o2, int dims, MATRIX *pmat){
//    OBJECT *p1 = oaget( poa, o1);
//    OBJECT *p2 = oaget( poa, o2);
//    NUMBER res2 = obdist2( p1, p2);

//    /* decrease it, by the projections */
//    int i;
//    for(i=0; i< dims; i++){
//	NUMBER temp = matrix2get(pmat, o1, i) - matrix2get(pmat, o2, i);
//	res2 -=  temp*temp;
//
//        //Modification:Heuristic  Wang et al.
//        if (res2<0) res2 *= -1;
//    }
//    return( res2);
//}

//NUMBER obdist2( OBJECT *p1, OBJECT *p2){
//    float res2; 
//    float v1,v2;
//    float  Sum1,Sum2;
//    int pos,length;
//
//    assert( (p1->len) == (p2->len) );
//    length = p1->len;
//    
//    Sum1 = 0.0;
//    Sum2 = 0.0; 
//    for(pos=0; pos<length; pos++){
//	v1 = (p1->vec)[pos];
//	v2 = (p2->vec)[pos];
//      Sum1 += v1;
//      Sum2 += v2; }
//
//    res2 = 0.0;
//    for(pos=0; pos<length; pos++){
//	v1 = (p1->vec)[pos];
//	v2 = (p2->vec)[pos];
//	res2 += sqrt((v1/Sum1)*(v2/Sum2)); }
//    
//    if (res2>1.0) res2 = 1.0;
//    res2 = 10000*(1-res2);
//
//    return(res2);
//}

/* reads in a matrix[0...n][0...Dim], into an object array */
OA *FastMap::Mat2oa(float** Mat,int n,int dim){
    OA *poa;
    OBJECT *pob;
    int k;

    poa=oaalloc(n);
    assert(poa != NULL);

    for(k=0;k<n;k++)
    { pob=Mat2ob(Mat,k,dim);
        oaput(poa,k,pob);
        obfree(pob); }

    return(poa);
}

/* reads each sample, into an object */
OBJECT *FastMap::Mat2ob(float** Mat,int numech,int dim){
    OBJECT *pob;
    int k;

    pob = oballoc(dim);
    for(k=0;k<dim;k++) obput(pob,k,Mat[numech][k]);
    return(pob);
}

/* convert a MATRIX into mat[0...n][0...dim] */
void FastMap::convertmat(MATRIX *pmatrix,float** Mat){
    int k,n;

    assert(pmatrix->count>0);
    n=pmatrix->count;
    for(k=0;k<n;k++) fillMat(matrixget(pmatrix,k),Mat,k);
}


void FastMap::fillMat(VECTOR *p,float** Mat,int n){
    int k,length;

    length=p->len;
    for(k=0;k<length;k++) Mat[n][k]=vecget(p,k);
}

/* FastMap  Matin[0...n][0...dim_in] >  Matout[0...n][0...dim_out] */
void FastMap::MyFastMap(float** Matin,float** Matout,int lgth,int dimin,int dimout,int verb) {
    MATRIX *mat;
    OA *a;

    a=Mat2oa(Matin,lgth,dimin);
    mat=fastmap(a,dimout);

    if (verb) matrixprint(mat);
    convertmat(mat,Matout);

    matrixfree(mat);
    oafree(a);
}

/* FastMap  Matin[0...n][0...dim_in] >  Matin[0...n][0...dim_out] */
void FastMap::MyFastMap_(float** Matin,int lgth,int dimin,int dimout,int verb) {
    MATRIX *mat;
    OA *a;

    a=Mat2oa(Matin,lgth,dimin);
    mat=fastmap(a,dimout);

    if (verb) matrixprint(mat);
    convertmat(mat,Matin);

    matrixfree(mat);
    oafree(a);
}

/*===================================================*/




/*===================================================*/
/*### Shifted Euclidean =============================*/
//## L 2
//
// NUMBER obdist2( OBJECT *p1, OBJECT *p2){
//    NUMBER res2; 
//    NUMBER resF = -1;
//    NUMBER v1, v2;
//    int pos, length, shift, idx;

//    assert( (p1->len) == (p2->len) );
//    length = p1->len;

//    // Shifted Euclidean Distance
//    for(shift = 0; shift < 66; shift+=6)
//       {
//        res2=0.0;
//        for(pos=0; pos< length; pos++)
//           {
//            v1 = (p1->vec)[pos];
//            idx = (shift + pos)%length;
//            v2 = (p2->vec)[idx];
//            res2 += (v1-v2)*(v1-v2);
//           }

//        // Find the min distance
//        if(resF == -1 || res2 < resF)
//            resF = res2;
//     }
//    return sqrt(resF);
// }



/*===================================================*/
/*### FFT ===========================================*/
void FastMap::fourn(float data[], unsigned long nn[], int ndim, int isign)
{
    int idim;
    unsigned long i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
    unsigned long ibit,k1,k2,n,nprev,nrem,ntot;
    float tempi,tempr;
    double theta,wi,wpi,wpr,wr,wtemp;

    for (ntot=1,idim=1;idim<=ndim;idim++)
        ntot *= nn[idim];
    nprev=1;
    for (idim=ndim;idim>=1;idim--) {
        n=nn[idim];
        nrem=ntot/(n*nprev);
        ip1=nprev << 1;
        ip2=ip1*n;
        ip3=ip2*nrem;
        i2rev=1;
        for (i2=1;i2<=ip2;i2+=ip1) {
            if (i2 < i2rev) {
                for (i1=i2;i1<=i2+ip1-2;i1+=2) {
                    for (i3=i1;i3<=ip3;i3+=ip2) {
                        i3rev=i2rev+i3-i2;
                        SWAP(data[i3],data[i3rev]);
                        SWAP(data[i3+1],data[i3rev+1]);
                    }
                }
            }
            ibit=ip2 >> 1;
            while (ibit >= ip1 && i2rev > ibit) {
                i2rev -= ibit;
                ibit >>= 1;
            }
            i2rev += ibit;
        }
        ifp1=ip1;
        while (ifp1 < ip2) {
            ifp2=ifp1 << 1;
            theta=isign*6.28318530717959/(ifp2/ip1);
            wtemp=sin(0.5*theta);
            wpr = -2.0*wtemp*wtemp;
            wpi=sin(theta);
            wr=1.0;
            wi=0.0;
            for (i3=1;i3<=ifp1;i3+=ip1) {
                for (i1=i3;i1<=i3+ip1-2;i1+=2) {
                    for (i2=i1;i2<=ip3;i2+=ifp2) {
                        k1=i2;
                        k2=k1+ifp1;
                        tempr=(float)wr*data[k2]-(float)wi*data[k2+1];
                        tempi=(float)wr*data[k2+1]+(float)wi*data[k2];
                        data[k2]=data[k1]-tempr;
                        data[k2+1]=data[k1+1]-tempi;
                        data[k1] += tempr;
                        data[k1+1] += tempi;
                    }
                }
                wr=(wtemp=wr)*wpr-wi*wpi+wr;
                wi=wi*wpr+wtemp*wpi+wi;
            }
            ifp1=ifp2;
        }
        nprev *= n;
    }
}

/*----------------------------------------------------------*/
/* FFT1D                                                    */
/*----------------------------------------------------------*/
void FastMap::FFT1D(float* vctR,float* vctI,int lgth)
{
    int i;
    float* data;
    unsigned long* nn;

    /*allocation memoire*/
    data=(float*)malloc(sizeof(float)*((2*lgth)+1));
    nn=(unsigned long*)malloc(sizeof(unsigned long)*2);
    nn[1]=lgth;

    /*Remplissage de data*/
    for(i=0;i<lgth;i++)
    { data[(2*i)+1]=vctR[i];
        data[(2*i)+2]=vctI[i];
    }

    /*FFTDD*/
    fourn(data,nn,1,1); // the last parameter is to select FFT

    /*Resultat*/
    for(i=0;i<lgth;i++)
    { vctR[i]=data[(2*i)+1];
        vctI[i]=data[(2*i)+2];  }

    /*Liberation memoire*/
    free(data);
    free(nn);
}

/*===================================================*/


/*===================================================*/
/*### Shifted Distance =============================*/
//
//
NUMBER FastMap::obdist2(OBJECT *p1, OBJECT *p2)
{
  NUMBER res2;
  NUMBER res_best; 
  NUMBER resF=-1;
  NUMBER v1,v2,v1p,v2p;
  int pos,length,shift,idx,idxp;

   assert( (p1->len) == (p2->len) );
   length = p1->len;

   //Non-Shifted L2 Dist.
   //-------------------
   if (MyShiftDistChoice==DIST_L2NS)
      { 
       resF=0.0;
       for(pos=0; pos<length; pos++)
          { v1 = (p1->vec)[pos];
            v2 = (p2->vec)[pos];
            resF += (v1-v2)*(v1-v2); }
       //resF/=length;
       return sqrt(resF); }

   //Shifted L2 Distance
   //-------------------
   if (MyShiftDistChoice==DIST_L2)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=0; pos<length; pos++)
                { v1 = (p1->vec)[pos];
                  idx = (shift + pos)%length;
                  v2 = (p2->vec)[idx];
                  res2 += (v1-v2)*(v1-v2); }
   	     //res2/=length;

              if (resF==-1 || res2<resF)  resF=res2; }
        return sqrt(resF); }

   // if (MyShiftDistChoice==DIST_L2)
   //    { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
   //         { res2=0.0;
   //           for(pos=0; pos<(length-SHIFT_DIST); pos++)
   //              { v1 = (p1->vec)[pos];
   //                idx = (shift + pos)%length;
   //                v2 = (p2->vec)[idx];
   //                res2 += (v1-v2)*(v1-v2); }
   //           // res2/=length; 

   //           if (resF==-1 || res2<resF)  resF=res2; }
   //      return sqrt(resF); }

   //Shifted L1 Distance
   //-------------------
   if (MyShiftDistChoice==DIST_L1)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=0; pos<length; pos++)
                { v1 = (p1->vec)[pos];
                  idx = (shift + pos)%length;
                  v2 = (p2->vec)[idx];
                  res2 += fabs(v1-v2); }
             //res2/=length;

             if (resF==-1 || res2<resF)  resF=res2; }
        return resF; }

   //Shifted LInf Distance
   //---------------------  
   if (MyShiftDistChoice==DIST_LInf)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=res_best=0.0;
             for(pos=0; pos<length; pos++)
                { v1 = (p1->vec)[pos];
                  idx = (shift + pos)%length;
                  v2 = (p2->vec)[idx];
                  res2 = (v1-v2)*(v1-v2); 
                  if (res2>res_best) res_best=res2; }
             //res2/=length; 

             if (resF==-1 || res_best<resF) resF=res_best; }
       return resF; }

   //LMod Distance 
   //---------------------  
   if (MyShiftDistChoice==DIST_LMod)
      { 
       int lgthP2=pow(2.0,(int)(log(length)/log(2)));

       NUMBER* re1 = new NUMBER[lgthP2];
       NUMBER* re2 = new NUMBER[lgthP2];
       NUMBER* im1 = new NUMBER[lgthP2];
       NUMBER* im2 = new NUMBER[lgthP2];
       NUMBER* mod1 = new NUMBER[lgthP2];
       NUMBER* mod2 = new NUMBER[lgthP2];

       for(int i=0; i<lgthP2; i++)
          { re1[i] = p1->vec[i];
            re2[i] = p2->vec[i]; 
            im1[i] = 0;
            im2[i] = 0; }

       FFT1D(re1,im1,lgthP2);
       FFT1D(re2,im2,lgthP2);
  
       for(int i=0; i<lgthP2; i++)
          { mod1[i] = sqrt(re1[i]*re1[i]+im1[i]*im1[i]); 
            mod2[i] = sqrt(re2[i]*re2[i]+im2[i]*im2[i]); }  

       //Distance L2 between module 
      resF=0.0;
      for(int pos=0;pos<lgthP2/DSTMODFCT;pos++)
         { v1 = mod1[pos];
           v2 = mod2[pos];
           resF += (v1-v2)*(v1-v2); }
      //resF /= lgthP2;
      free(re1); free(re2);   free(im1); 
      free(im2); free(mod1); free(mod2);
      return sqrt(resF); 
      }

   //LRad Distance 
   //---------------------  
   if (MyShiftDistChoice==DIST_LRad)
      {
       int n_bins = 255;
       int bin_w = int(255/n_bins); 
       bin_w = (bin_w <= 1.0)? 1 : bin_w+1;      // +1 to ensure that the max index < n_bins
       int* h1 = new int[n_bins];
       int* h2 = new int[n_bins];

       for(int i=0; i<n_bins; i++) { h1[i]=0; h2[i]=0; }

       int idx;
       for(int pos=0; pos<length; pos++)
          { idx = (int) (p1->vec[pos])/bin_w;
            if (idx<n_bins) h1[idx]++; else h1[n_bins-1]++;
            idx = (int) (p2->vec[pos])/bin_w;
            if (idx < n_bins) h2[idx]++; else h2[n_bins-1]++; }
 
      resF=0;
      for(int i=0; i< n_bins; i++)
         { v1 = h1[i];
           v2 = h2[i];
           resF += (v1-v2)*(v1-v2); }

      free(h1); free(h2);
      return sqrt(resF);
      }

   //Shifted LGrd Distance 
   //--------------------- 
   if (MyShiftDistChoice==DIST_LGrd)
      { 
        float MoyV1=0,MoyV2=0;
        const int DIFGRD=4;

        for(pos=0; pos<length; pos++) 
	   { MoyV1+=(p1->vec)[pos];
             MoyV2+=(p2->vec)[pos]; }

        MoyV1/=length;
        MoyV2/=length;

        for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=DIFGRD; pos<(length - DIFGRD); pos++)
                { v1 = (p1->vec)[pos - DIFGRD];
                  v1p = (p1->vec)[pos + DIFGRD];
                  idx = (shift + pos - DIFGRD)%length;
                  idxp = (shift + pos + DIFGRD)%length;
                  v2 = (p2->vec)[idx];
                  v2p = (p2->vec)[idxp];
                  res2 += fabs((v1-v1p)-(v2-v2p)); }
             //res2/=length;

             if (resF==-1 || res2<resF)  resF=res2; }
        return resF; }

   // LMoy Distance 
   //--------------------- 
    if (MyShiftDistChoice==DIST_LMoy)
       {
	float MoyV1=0,MoyV2=0;
  
        for(pos=0; pos<length; pos++) 
	   { MoyV1+=(p1->vec)[pos];
             MoyV2+=(p2->vec)[pos]; }

        resF=fabs(MoyV1-MoyV2);
        //resF/=length;
        return resF; }

   // TEST::[1] 
   //--------------------- 
   if (MyShiftDistChoice==8)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=0; pos<(length-SHIFT_DIST); pos++)
                { v1 = (p1->vec)[pos];
                  idx = (shift + pos)%length;
                  v2 = (p2->vec)[idx];
                  res2 += (v1-v2)*(v1-v2); }
             // res2/=length; 

             if (resF==-1 || res2<resF)  resF=res2; }
        return sqrt(resF); }

   // TEST::[2] 
   //--------------------- 
   if (MyShiftDistChoice==9)
      { for(shift=0; shift<30; shift+=2)
           { res2=0.0;
             for(pos=0; pos<(length-30); pos++)
                { v1 = (p1->vec)[pos];
                  idx = (shift + pos)%length;
                  v2 = (p2->vec)[idx];
                  res2 += (v1-v2)*(v1-v2); }
             // res2/=length; 

             if (resF==-1 || res2<resF)  resF=res2; }
        return sqrt(resF); }

   // TEST::[3] 
   //--------------------- 
   if (MyShiftDistChoice==100)
      { v1 = (p1->vec)[0];
        v2 = (p2->vec)[0];
        resF = (v1-v2)*(v1-v2);
        return sqrt(resF); }

 //Return
 return resF;
}



