//---------------------------------------------------
// module  : f_MdsBasedSymEst.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "MdsBasedSymEst.h"

//----------------------------------------------------------
// Constructeur de la classe MdsBasedSymEst
//----------------------------------------------------------
MdsBasedSymEst::MdsBasedSymEst(char* Nm,unsigned short*** DpthMpSq,int lgth,int wdth,int dpth)      
{
 int i,j,k;

 //RecordParameters
 strcpy(Name,Nm);
 depth=dpth;
 height=lgth;
 width=wdth;

 //DefaultParameters
 FactPrior=FACT_PRIOR;

 //InitParameters
 NrjBeg=NrjEnd=0.0;
 CorelRaw=CorelRefined=0.0;
 ASIndx=0.0;
 PourcPixSatRGB=0.0;

 //Memory_Allocation
 DepthMapSeq=usimatrix_allocate_3d(dpth,lgth,wdth);
 RawMdsSymMap=fmatrix_allocate_3d(TROIS,lgth,wdth);
 ProcMdsSymMap=fmatrix_allocate_3d(TROIS,lgth,wdth);
 RGBMdsSymMap=fmatrix_allocate_3d(TROIS,lgth,wdth);
 ASIcurve=fmatrix_allocate_1d(height);

 //Record_DepthMapSeq
 for(k=0;k<depth;++k)
 for(i=0;i<height;++i) for(j=0;j<width;++j)
 DepthMapSeq[k][i][j]=DpthMpSq[k][i][j];
}

//----------------------------------------------------------
// Destructeur de la classe MdsBasedSymEst
//----------------------------------------------------------
MdsBasedSymEst::~MdsBasedSymEst()
{ 
 //Libere Memoire Matrice
 if (DepthMapSeq)   free_usimatrix_3d(DepthMapSeq,depth);
 if (RawMdsSymMap)  free_fmatrix_3d(RawMdsSymMap,TROIS);  
 if (ProcMdsSymMap) free_fmatrix_3d(ProcMdsSymMap,TROIS);
 if (RGBMdsSymMap)  free_fmatrix_3d(RGBMdsSymMap,TROIS);
 if (ASIcurve)      free_fmatrix_1d(ASIcurve);
}

//------------------------------------------------
// METHODES DE LA CLASSE MdsBasedSymEst
//------------------------------------------------
//------------------------------------------------
// MdsRoughEstimation
//------------------------------------------------
void MdsBasedSymEst::MdsRoughEstimation(int dst,unsigned short*** ImgSym)
{
 int i,j,k;
 int h,w;

 //RecordMdsModelParameters
 DistMap=dst;

 //MemoryAllocation
 float** matx=fmatrix_allocate_2d((height*width),depth);
 float*** mattmp=fmatrix_allocate_3d(TROIS,height,width);

 //Initialization
 for(k=0;k<depth;++k)
 for(i=0;i<height;++i) for(j=0;j<width;++j)
 matx[i*width+j][k]=DepthMapSeq[k][i][j];

 //Presentation
 printf(" : Start FastMap(dst=%d|sft=%d|stp=%d])...",DistMap,SHIFT_DIST,STEP_SHIFT);
 fflush(stdout);
 FastMap* fmp;
 fmp=new FastMap;
 (*fmp).MyShiftDistChoice=DistMap; 
 (*fmp).MyFastMap_(matx,(height*width),depth,TROIS,0);
 printf("done.");
 fflush(stdout);

 //MDS-SymMap Result
 for(k=0;k<(height*width);k++)
    { h=(int)(k/width);
      w=(int)(k%width);

      mattmp[0][h][w]=matx[k][0];
      mattmp[1][h][w]=matx[k][1];
      mattmp[2][h][w]=matx[k][2]; }
    
 //LinearStretch
 RecalB(mattmp,height,width); 
 
 //Transfert
 for(k=0;k<TROIS;++k)
 for(i=0;i<height;++i) for(j=0;j<width;++j)
    { RawMdsSymMap[k][i][j]=(unsigned short)mattmp[k][i][j];
      ProcMdsSymMap[k][i][j]=RawMdsSymMap[k][i][j];
      ImgSym[k][i][j]=RawMdsSymMap[k][i][j]; }

 //RecordInfo
 FillInfoNameDist();

 //FreeMemory_&_Object
 if (matx)   free_fmatrix_2d(matx);
 if (mattmp) free_fmatrix_3d(mattmp,TROIS);
 if (fmp)    delete(fmp);
}

//------------------------------------------------
// IterRefineEstimation
//------------------------------------------------
void MdsBasedSymEst::ComputeNonLocalGraph(int NbCnx,int ExplrWinSz)
{
 int i,j,k;

 //Record
 NumbCnx=NbCnx;
 ExploreWinSize=ExplrWinSz;

 //AllocateMemoryNonLocalGraph
 NonLocalGraph=usimatrix_allocate_3d(TROIS,NbCnx,height*width);

 //Initialization_NonLocalGraph_[3][nconnexion][num_pixel]
 for(j=0;j<height;++j) for(i=0;i<width;++i)
 for(k=0;k<NbCnx;++k)
    { NonLocalGraph[0][k][j*width+i]=j; //Row
      NonLocalGraph[1][k][j*width+i]=i; //Col
      NonLocalGraph[2][k][j*width+i]=0; }
        
 //Non-Local Graph Evaluation
 Compute_NonLocalGraph(DepthMapSeq,depth,NonLocalGraph,height,width,ExplrWinSz,NbCnx);

 //Debug
 short int max=0;
 for(j=0;j<height*width;j++) for(k=0; k<NbCnx;k++)
 if(max<NonLocalGraph[2][k][j]) max=NonLocalGraph[2][k][j];
 printf(" (AfterNorm.:DistMx=%d)...done\n",max);
}

//------------------------------------------------
// MetropolisRefine
// ----------------
//
// scalingfact*dist(NonLocGrph)-dist(SymMap)
// dist(NonLocGrph)-dist(SymMap)/scalingfact
// dist(NonLocGrph)-dist(RawMdsSymMap)/scalingfact
//------------------------------------------------
void MdsBasedSymEst::MetropolisRefine(unsigned short*** ResultSymMap,float temp0,float tempfinal,float radius,int maxiter)
{ 
 int i,j,k,l,d;
 float energy,energy_old,energy_new,delta_e;
 float nchanges,newval; 
 float prob;

 //Constant
 const int ITERINFO=10;   

 //Stretch  NonLocalGrph [DepthMapSeq - RawMdsSymMap>ProcMdsSymMap]
 ScalingFact=Compute_Scaling_Factor(); 
 ScaleDistNonLocalGrph(ScalingFact);

 for(j=0;j<height;j++) for(i=0;i<width;i++) for(k=0;k<TROIS;k++)
 ProcMdsSymMap[k][j][i]=((float)RawMdsSymMap[k][j][i]);

 //Info
 Stat(ProcMdsSymMap[0],height,width);
 Stat(ProcMdsSymMap[1],height,width);
 Stat(ProcMdsSymMap[2],height,width);

    //Arrays allocation and initialization
    float**** matrixgrid = fmatrix_allocate_4d(2,TROIS,height,width);
    for(l=0;l<TROIS;l++) CopyMatrix(ProcMdsSymMap[l],matrixgrid[0][l],height,width);
    for(l=0;l<TROIS;l++) CopyMatrix(ProcMdsSymMap[l],matrixgrid[1][l],height,width);

    float* xs=fmatrix_allocate_1d(TROIS);
    float* ys=fmatrix_allocate_1d(TROIS);

    //Temperature
    float tratio=powf((tempfinal/temp0),1.0/((float)maxiter)); 
    float temp=temp0*tratio;
    int it;

    //Record
    NrjBeg=Compute_Global_Energy(matrixgrid[0],1.0,0.0);
    energy=NrjBeg;

   //=[LOOP-METROPOLI]================================
    printf("\n > Start Local Search...[FactPrior=%.3f][Stochast=%d]",FactPrior,FlgStoch);
    for(it=0; it<maxiter && temp > tempfinal; ++it, temp=temp0*powf(tratio,it)) 
       {
         if ((it)%ITERINFO==0) for(l=0;l<TROIS;l++) CopyMatrix(matrixgrid[0][l],matrixgrid[1][l],height,width);
        
          for(j=PATIENT_YMIN;j<PATIENT_YMAX;j++) for(i=PATIENT_XMIN;i<PATIENT_XMAX;i++)
          for(d=0;d<TROIS;d++)
             {
                //Sample a new value
                newval = matrixgrid[0][d][j][i] - radius/2 + radius*randomize() + 0.49999;

                if (newval<0) newval=0.0;
                for(k=0;k<TROIS;k++)   { ys[k]=matrixgrid[0][k][j][i];  xs[k]=matrixgrid[0][k][j][i]; }
                ys[d]=newval; 

                //Compute_Local_Energies
                energy_old = Compute_Local_Energy(matrixgrid[0], j, i, xs, 1.0,FactPrior); //FACT_PRIOR
                energy_new = Compute_Local_Energy(matrixgrid[0], j, i, ys, 1.0,FactPrior);
                delta_e = (energy_new - energy_old);

                //Test whether to update the matrix and replace its value
                if (!FlgStoch) if (delta_e<0)  matrixgrid[0][d][j][i]=newval;

                if (FlgStoch)
		   { if (delta_e<0) matrixgrid[0][d][j][i]=newval;
                     else { prob = (exp(-(delta_e)/temp));
                            if (prob>randomize())  { matrixgrid[0][d][j][i]=newval; } } }
	     }

        //Print Info
	if (it%ITERINFO==0)  { energy = Compute_Global_Energy(matrixgrid[0],1.0,0.0);
                               nchanges = DiffMat(matrixgrid[1][0],matrixgrid[0][0],height,width,PREC_MET);
                               //printf("\n %d %.3f",it,energy);
                               printf("\n     [%d]>%.3f [Temp:%.4f][NbChgt/100=%.2f]",it,energy,temp,nchanges); 
                              }
       }//=EndLoop=======================

    //Copy Final Solution
    for(l=0;l<TROIS;l++) CopyMatrix(matrixgrid[0][l],ProcMdsSymMap[l],height,width);
    NrjEnd=energy;
    printf("\n > ...Local Refining done. [Nrj=%.3f>%.3f]",NrjBeg,NrjEnd);

    //Stat
    Stat(ProcMdsSymMap[0],height,width);
    Stat(ProcMdsSymMap[1],height,width);
    Stat(ProcMdsSymMap[2],height,width);
    RecalB(ProcMdsSymMap,height,width);
    Stat(ProcMdsSymMap[0],height,width);
    Stat(ProcMdsSymMap[1],height,width);
    Stat(ProcMdsSymMap[2],height,width);
    printf("\n");

    //Free Memory
    if(matrixgrid) free_fmatrix_4d(matrixgrid,2,TROIS);
    if(xs) free_fmatrix_1d(xs);
    if(ys) free_fmatrix_1d(ys);

    //Transfert>Result
    for(k=0;k<TROIS;++k) for(i=0;i<height;++i) for(j=0;j<width;++j) 
    ResultSymMap[k][i][j]=ProcMdsSymMap[k][i][j];
}

//------------------------------------------------
// ColorConvLABtoRGBNonSat
//------------------------------------------------
void MdsBasedSymEst::ColorConvLABtoRGBNonSat(unsigned short*** RGBSymMapRes)
{ 
 int i,j,k;

  //Find GoodRange For LAB
  Set_Good_Range_LAB();

  //Conversion
  PourcPixSatRGB=Convert_LAB_TO_RGB(ProcMdsSymMap,height,width,RGBMdsSymMap);
  
  //LinearStretch
  RecalB(RGBMdsSymMap,height,width);

  //Info
  printf("\n > Stat.Img.RGB");
  Stat(RGBMdsSymMap[0],height,width);
  Stat(RGBMdsSymMap[1],height,width);
  Stat(RGBMdsSymMap[2],height,width);

  //Transfert>Result
  for(k=0;k<TROIS;++k) for(i=0;i<height;++i) for(j=0;j<width;++j) 
  RGBSymMapRes[k][i][j]=RGBMdsSymMap[k][i][j];
}


//-----------------------------
//-----------------------------
// GESTION --------------------
//-----------------------------
//-----------------------------
//------------------------------------------------
// Compute_NonLocalGraph_
//------------------------------------------------
void MdsBasedSymEst::Compute_NonLocalGraph(unsigned short*** seq_in,int depth,\
unsigned short*** nonlocgraph,int height,int width,int sizewin,int nconnexion)
{
 int i,j,k,l,d;
 int nnn=4;                                             //Nb nearest neighbours 
 int numel=sizewin*sizewin;                             //Nb element for local search
 int sizediv=(int)(numel/((float)(nconnexion-nnn)))+1;  //Size of space between neighbours
 int c_n=0;                                             //Counter for the nb of node in the graph
 int posr,posc;                                         //Absolute position of the node in image
 float dst;
 float** distances;
  
 //AllocationMemory
 float* p1=new float[depth];
 float* p2=new float[depth];
 distances=fmatrix_allocate_2d(nconnexion,height*width);

    //------------------------------------------------
    //--- Compute the graph!
    //------------------------------------------------
    for(j=0;j<height;++j) for(i=0;i<width;++i)
       {
        //Init vector at current pixel position
	 for(d=0;d<depth;++d) p1[d]=seq_in[d][j][i];
        c_n=0; //neighbours counter

        //Big windows search
        for(k=0;k<sizewin;++k) for(l=0;l<sizewin;++l)
               {
                // Each node is equally spaced by sizediv elements
                // therefore if the element is not a multiple of sizediv
                // the search process continue

                if((k*sizewin + l) % sizediv)   continue;

                //Else let fill the graph
                posr = j - sizewin/2 + k;
                posc = i - sizewin/2 + l;
                if (posr<0)             posr+=height;
                if (posr>(height-1))    posr-=height;
                if (posc<0)             posc+=width;
                if (posc>(width-1))     posc-=width;

                for(d=0;d<depth;d++) p2[d]=seq_in[d][posr][posc]; //neighbour
	        dst=Myobdist(p1,p2,depth); //dist

                //Store infos
                nonlocgraph[0][c_n][j*width+i]= posr; //Row
                nonlocgraph[1][c_n][j*width+i]= posc; //Col
                distances[c_n][j*width+i]     = dst;  //Distance
                ++c_n;
               }


        //4-Nearest Neighours
        for(k=-1;k<=1;k++) for(l=-1;l<=1;l++)
           { //relative pos of neighbours
             if((k!=0 && l==0) || (k==0 && l!=0))
               {
                posr = j + k;
                posc = i + l;
                if (posr<0)             posr+=height;
                if (posr>(height-1))    posr-=height;
                if (posc<0)             posc+=width;
                if (posc>(width-1))     posc-=width;

                for(d=0;d<depth;++d) p2[d]=seq_in[d][posr][posc]; //neighbour
                dst=Myobdist(p1,p2,depth); //dist

                //Store infos
                nonlocgraph[0][c_n][j*width+i]= posr; //Row
                nonlocgraph[1][c_n][j*width+i]= posc; //Col
                distances[c_n][j*width+i]     = dst ; //Distance
                ++c_n;
	       }
           }
       }//--EndGraph----------------------------------------------

    //Normalize distances and assign them to the graph
    float max,min;
    max=min=distances[0][0];
    for(i=0;i<height*width;i++) for(k=0; k<nconnexion; k++)
       { if(max < distances[k][i]) max=distances[k][i];
         if(min > distances[k][i]) min=distances[k][i]; }
    //printf(" min=%.4f",min);

    for(i=0;i<height*width;i++) for(k=0;k<nconnexion;k++)
    distances[k][i] *= 500.0/max;  

    for(j=0;j<height;++j) for(i=0;i<width;++i)
    for(k=0;k<nconnexion;k++)
    nonlocgraph[2][k][j*width +i] = (short int)distances[k][j*width+i];
    
    //Info
    printf("\n   (NonLocGrph[min=%.1f::max=%.1f])",min,max);
}

//------------------------------------------------
// ScaleDistNonLocalGrph
//------------------------------------------------
void MdsBasedSymEst::ScaleDistNonLocalGrph(float scale)
{
 int i,j,k;
 float tmp,min,max;

 //ScaleGrph
 min=max=NonLocalGraph[2][0][0];
 for(j=0;j<height;j++) for(i=0;i<width;i++) 
 for(k=0;k<NumbCnx;k++)
    { tmp=NonLocalGraph[2][k][j*width+i];
      tmp*=scale;
      if (tmp>max) max=tmp;
      if (tmp<min) min=tmp;
      NonLocalGraph[2][k][j*width+i] = (short int)(tmp+0.5); }
 printf("   [min=%.2f][max=%.2f]",min,max);
}

//------------------------------------------------
// Compute_Scaling_Factor
//
// scalingfac*dist(NonLocGrph)-dist(SymMap)
//------------------------------------------------
float MdsBasedSymEst::Compute_Scaling_Factor()
{
 int i;
 float kopt,scalingfac,scalingfactorInit;
 float energytemp,energyopt;

 //Init with (default init value)
 scalingfactorInit=scalingfac=kopt=0.6;
 energytemp=energyopt=Compute_Global_Energy(RawMdsSymMap,scalingfac,0.0);

 //Estimate_Scalingfact
 for(i=0;scalingfac>0.0;i++)
    { scalingfac -= 0.005;
      energytemp = Compute_Global_Energy(RawMdsSymMap,scalingfac,0.0);

      //DEBUG
      //printf("[%.3f::%.1f]",scalingfac,energytemp);

      if (energytemp<energyopt)
         { energyopt=energytemp;
           kopt=scalingfac; }
    }

  //DEBUG
  if (kopt==scalingfactorInit) printf("\n PROBLEME!!!!!!!(scalingfac superieur !!");

  //Assign the best scaling factor & Return
  ScalingFact=kopt;
  printf(" > Scaling factor:[%.3f>>%.3f] For Nrj:[%.2f]\n",scalingfactorInit,kopt,energyopt);
  return kopt;
}

//------------------------------------------------
// Compute_Global_Energy
//------------------------------------------------
float MdsBasedSymEst::Compute_Global_Energy(float*** SymMap,float scalingfactor,float fact_prior)
{
 int i,j,k;
 float energy = 0.0;
 float us[TROIS];

 //Loop_For_Each_Pixel
 for(j=0;j<height;j++) for(i=0;i<width;i++)
    { 
     for(k=0;k<TROIS;k++) us[k]=SymMap[k][j][i];
     energy += Compute_Local_Energy(SymMap,j,i,us,scalingfactor,fact_prior);
    }

 energy /= (width*height);
 return (float)energy;
}

//------------------------------------------------
// Compute_local_Energy
//------------------------------------------------
float MdsBasedSymEst::Compute_Local_Energy(float*** SymMap,int row,int col,float* val,\
float scalingfac,float prior_fact)
{
 int i,k,l;
 float energy;
 float likelihood, prior;

 //Pixels values in projected space
 float us[TROIS];
 float ut[TROIS];
 int posr,posc; 

 //Current pixel values in projected space
 us[0]=val[0];
 us[1]=val[1];
 us[2]=val[2];

 //Compute Energy for local neighbours 
 //===================================
 // Likelihood
 //-----------
 float term1;
 float odist, pdist;  //Original and Projected Distance 
 unsigned int numpix; //the ith pixel
 likelihood=0.0;
 odist=pdist=0.0;

 numpix=(row*width)+col;
 for(k=0;k<NumbCnx;k++)
    {
     posr = NonLocalGraph[0][k][numpix];
     posc = NonLocalGraph[1][k][numpix];
     odist = NonLocalGraph[2][k][numpix];

     ut[0] = SymMap[0][posr][posc];
     ut[1] = SymMap[1][posr][posc];
     ut[2] = SymMap[2][posr][posc];

     term1 = 0.0;
     for(i=0;i<TROIS;i++) term1 += (us[i]-ut[i])*(us[i]-ut[i]);
     pdist = sqrt(term1);

     likelihood += (scalingfac*odist-pdist)*(scalingfac*odist-pdist);
    }

    //GG_MRF Prior 
    //-----------------------
    float beta;
    float dist=0.0;
    prior = 0.0;
        //8_Nearest_Neighours
        for(k=-1; k<=1; k++) for(l=-1; l<=1; l++)
           {
            posr = row + k;
            posc = col + l;
            if (posr<0)             posr+=height;
            if (posr>(height-1))    posr-=height;
            if (posc<0)             posc+=width;
            if (posc>(width-1))     posc-=width;

            ut[0] = SymMap[0][posr][posc];
            ut[1] = SymMap[1][posr][posc];
            ut[2] = SymMap[2][posr][posc];

            //Relative Neighbours Pos.
            if((k!=0 && l==0) || (k==0 && l!=0))    beta=1.0;           
            else if (k != 0 && l!=0)                beta=0.70710678;    
            else continue;                                              

            for(i=0;i<TROIS;++i) dist += pow(fabs(us[i]-ut[i]), Q_PRIOR);
            prior+= beta*dist;
          }

    energy = likelihood/NumbCnx + prior_fact*prior/(8.0);
    return energy;
}

//------------------------------------------------
// Set_Good_Range_LAB
//------------------------------------------------
void MdsBasedSymEst::Set_Good_Range_LAB()
{ 
 int i,j;

 //Presentation
 printf("\n > SetGoodRangeFor_LAB: Before:");
 
  //Stat_LAB
  Stat(ProcMdsSymMap[0],height,width);
  Stat(ProcMdsSymMap[1],height,width);
  Stat(ProcMdsSymMap[2],height,width);

  //Linear_Stretch
  RecalB(ProcMdsSymMap,height,width);

  //LAB_Range
  RecalI(ProcMdsSymMap[0],height,width,0,100);
  RecalI(ProcMdsSymMap[1],height,width,0,100);
  RecalI(ProcMdsSymMap[2],height,width,0,100);

 //Center a* and b* channels
 float mean1 = ComputeMean(ProcMdsSymMap[1],height,width);
 for(j=0;j<height;j++) for(i=0;i<width;i++) ProcMdsSymMap[1][j][i] -= mean1;

 float mean2 = ComputeMean(ProcMdsSymMap[2],height,width);
 for(j=0;j<height;j++) for(i=0;i<width;i++) ProcMdsSymMap[2][j][i] -= mean2;

  //Stat LAB
  Stat(ProcMdsSymMap[0],height,width);
  Stat(ProcMdsSymMap[1],height,width);
  Stat(ProcMdsSymMap[2],height,width);
}

//------------------------------------------------
// ASI Index
//------------------------------------------------
float MdsBasedSymEst::EstimateASI(float*** SymMap)
{
 int i,j,k,l,m,p;
 int nb;
 float tmp,nrj;
 float ASI;

 int ValMiddle=0;
 int PatientYmin=0;
 int PatientYmax=0;
 int pos=0;

 //Cst
 const int MIN_GRAD=100;
 //const int STEP_FILE=4;
 const int HeightBegin=(int)(0.250*(float)height);
 const int HeightEnd=(int)(0.875*(float)height);
 const int WidthDiv2=width/2.0;
 const int IntMidle=10;
 for(nb=0,l=HeightBegin;l<HeightEnd;l+=4) nb++;
 
 //Presentation
 printf("\n > Estimate ASI...");
 fflush(stdout);

 //MemoryAllocation
 float** SymMapGrd=fmatrix_allocate_2d(height,width);
 unsigned short* VectMiddle=usimatrix_allocate_1d(nb);
 float* VectASIfunct=fmatrix_allocate_1d(height);

 //Initialization
 for(i=0;i<height;++i) for(j=0;j<width;++j) SymMapGrd[i][j]=0.0;
 for(i=0;i<height;++i) VectASIfunct[i]=0.0;

 //Grad-Estimate_(L1 Norm)
 for(i=1;i<(height-1);++i) for(j=1;j<(width-1);++j)
    { tmp=fabs(SymMap[0][i-1][j]-SymMap[0][i+1][j]);
      tmp+=fabs(SymMap[0][i][j-1]-SymMap[0][i][j+1]);
      tmp+=fabs(SymMap[1][i-1][j]-SymMap[1][i+1][j]);
      tmp+=fabs(SymMap[1][i][j-1]-SymMap[1][i][j+1]);
      tmp+=fabs(SymMap[2][i-1][j]-SymMap[2][i+1][j]);
      tmp+=fabs(SymMap[2][i][j-1]-SymMap[2][i][j+1]);
      SymMapGrd[i][j]=tmp; }
 Recal(SymMapGrd,height,width);
  
 //Middle_Estimate------------------------------
 p=-1;
 for(l=HeightBegin;l<HeightEnd;l+=4)
    { //LineSearch
      nrj=0.0;
      for(k=(WidthDiv2-IntMidle);k<(WidthDiv2+IntMidle);k++)
      for(m=0;m<WidthDiv2;m++)
	 { tmp=SymMapGrd[l][(k-m)%width]+SymMapGrd[l][(k+m)%width];
	   if (tmp>nrj) { nrj=tmp; pos=k; } }  
      //Record
      VectMiddle[++p]=(unsigned short)pos;
     }

 
 insertion_sort(VectMiddle,nb);
 ValMiddle=VectMiddle[nb/2];
 printf("[mid:%d]",ValMiddle);
 //-End-Middle Estimate---------------------

 //Patient_Ymin/Ymax_Estimate----------------
 for(l=0;l<height;l++)
    { for(tmp=0.0,k=0;k<width;k++)  tmp+=SymMapGrd[l][k];
      if (tmp>MIN_GRAD) { PatientYmin=l; break; } }

 for(l=(height-1);l>0;l--)
    { for(tmp=0.0,k=0;k<width;k++)  tmp+=SymMapGrd[l][k];
      if (tmp>MIN_GRAD) { PatientYmax=l; break; } } 
 
 printf("[Ymin/max::%d|%d]",PatientYmin,PatientYmax);
 ///-End-Ymin/Ymax_Estimate------------------

 //ASI_Estimate------------------------------
 for(l=PatientYmin;l<PatientYmax;l++)
    { nrj=0.0;
      for(k=0;k<WidthDiv2;k++)
         { tmp=CARRE(SymMap[0][l][(ValMiddle-k)%width]-SymMap[0][l][(ValMiddle+k)%width]);
	   tmp+=CARRE(SymMap[1][l][(ValMiddle-k)%width]-SymMap[1][l][(ValMiddle+k)%width]);
	   tmp+=CARRE(SymMap[2][l][(ValMiddle-k)%width]-SymMap[2][l][(ValMiddle+k)%width]); 
           tmp=sqrt(tmp);
           if (tmp>nrj)  nrj=tmp; }

       VectASIfunct[l]=nrj; }

  nrj=0.0;
  for(l=PatientYmin;l<PatientYmax;l++) nrj+=VectASIfunct[l];
  ASI=nrj/(PatientYmax-PatientYmin);
  printf(">ASI=%.2f",ASI);
 ///-End-ASI_Estimate------------------------

 //Save ASI_curve
 //FILE *fileASI=fopen("FileASI.dat","w");
 //for(l=0;l<height;l+=4) fprintf(fileASI,"%d %d\n",l/STEP_FILE,(int)VectASIfunct[l]);
 //for(l=0;l<height;l++) 
 //   { fprintf(fileASI,"%d %d\n",l,(int)VectASIfunct[l]);
 //     ASIcurve[l]=VectASIfunct[l]; } 
 //fclose(fileASI);

 //Save ASI_curve
 for(l=0;l<height;l++) ASIcurve[l]=VectASIfunct[l]; 

 //FreeMemory
 if (SymMapGrd)    free_fmatrix_2d(SymMapGrd);
 if (VectMiddle)   free_usimatrix_1d(VectMiddle);
 if (VectASIfunct) free_fmatrix_1d(VectASIfunct);

 //Record
 ASIndx=ASI;
 sprintf(Info,"_ASI_%.3f",ASIndx);

 //Return
 return ASI;
}

//-----------------------------
//-----------------------------
// DIST -----------------------
//-----------------------------
//-----------------------------

float MdsBasedSymEst::Myobdist(float* p1,float* p2,int length)
{
  float res2;
  float res_best; 
  float resF=-1;
  float v1,v2,v1p,v2p;
  int pos,shift,idx,idxp;

   //Non-Shifted L2 Dist.
   //-------------------
   if (DistMap==DIST_L2NS)
      { 
       resF=0.0;
       for(pos=0; pos<length; pos++)
          { v1 = p1[pos];
            v2 = p2[pos];
            resF += (v1-v2)*(v1-v2); }
       //resF/=length;
       return sqrt(resF); }

   //Shifted L2 Distance
   //-------------------
   if (DistMap==DIST_L2)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=0; pos<length; pos++)
                { v1 = p1[pos];
                  idx = (shift + pos)%length;
                  v2 = p2[idx];
                  res2 += (v1-v2)*(v1-v2); }
   	     //res2/=length;

              if (resF==-1 || res2<resF)  resF=res2; }
        return sqrt(resF); }

   // if (DistMap==DIST_L2)
   //    { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
   //         { res2=0.0;
   //           for(pos=0; pos<(length-SHIFT_DIST); pos++)
   //              { v1 = p1[pos];
   //                idx = (shift + pos)%length;
   //                v2 = p2[idx];
   //                res2 += (v1-v2)*(v1-v2); }
   // 	     //res2/=length; 

   //           if (resF==-1 || res2<resF)  resF=res2; }
   //      return sqrt(resF); }

   //Shifted L1 Distance
   //-------------------
   if (DistMap==DIST_L1)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=0; pos<length; pos++)
                { v1 = p1[pos];
                  idx = (shift + pos)%length;
                  v2 = p2[idx];
                  res2 += fabs(v1-v2); }
             //res2/=length;

             if (resF==-1 || res2<resF)  resF=res2; }
        return resF; }

   //Shifted LInf Distance
   //---------------------  
   if (DistMap==DIST_LInf)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=res_best=0.0;
             for(pos=0; pos<length; pos++)
                { v1 = p1[pos];
                  idx = (shift + pos)%length;
                  v2 = p2[idx];
                  res2 = (v1-v2)*(v1-v2); 
                  if (res2>res_best) res_best=res2; }
             //res2/=length; 

             if (resF==-1 || res_best<resF) resF=res_best; }
       return resF; }

   //LMod Distance 
   //---------------------  
   if (DistMap==DIST_LMod)
      { 
       int lgthP2=pow(2.0,(int)(log(length)/log(2)));

       float* re1 = new float[lgthP2];
       float* re2 = new float[lgthP2];
       float* im1 = new float[lgthP2];
       float* im2 = new float[lgthP2];
       float* mod1 = new float[lgthP2];
       float* mod2 = new float[lgthP2];

       for(int i=0; i<lgthP2; i++)
          { re1[i] = p1[i];
            re2[i] = p2[i]; 
            im1[i] = 0;
            im2[i] = 0; }

       MyFFT1D(re1,im1,lgthP2);
       MyFFT1D(re2,im2,lgthP2);
  
       for(int i=0; i<lgthP2; i++)
          { mod1[i] = sqrt(re1[i]*re1[i]+im1[i]*im1[i]); 
            mod2[i] = sqrt(re2[i]*re2[i]+im2[i]*im2[i]); }  

       //Distance L2 between module 
      resF=0.0;
      for(int pos=0;pos<lgthP2/DSTMODFCT;pos++)
         { v1 = mod1[pos];
           v2 = mod2[pos];
           resF += (v1-v2)*(v1-v2); }
      //resF /= lgthP2;
      free(re1); free(re2);   free(im1); 
      free(im2); free(mod1); free(mod2);
      return sqrt(resF); 
      }

   //LRad Distance 
   //---------------------  
   if (DistMap==DIST_LRad)
      {
       int n_bins = 255;
       int bin_w = int(255/n_bins); 
       bin_w = (bin_w <= 1.0)? 1 : bin_w+1;      
       int* h1 = new int[n_bins];
       int* h2 = new int[n_bins];

       for(int i=0; i<n_bins; i++) { h1[i]=0; h2[i]=0; }

       int idx;
       for(int pos=0; pos<length; pos++)
          { idx = (int) (p1[pos])/bin_w;
            if (idx<n_bins) h1[idx]++; else h1[n_bins-1]++;
            idx = (int) (p2[pos])/bin_w;
            if (idx < n_bins) h2[idx]++; else h2[n_bins-1]++; }
 
      resF=0;
      for(int i=0; i< n_bins; i++)
         { v1 = h1[i];
           v2 = h2[i];
           resF += (v1-v2)*(v1-v2); }

      free(h1); free(h2);
      return sqrt(resF);
      }

   //Shifted LGrd Distance 
   //--------------------- 
   if (DistMap==DIST_LGrd)
      { 
        float MoyV1=0,MoyV2=0;
        const int DIFGRD=4;

        for(pos=0; pos<length; pos++) 
	   { MoyV1+=p1[pos];
             MoyV2+=p2[pos]; }

        MoyV1/=length;
        MoyV2/=length;

        for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=DIFGRD; pos<(length - DIFGRD); pos++)
                { v1 = p1[pos - DIFGRD];
                  v1p = p1[pos + DIFGRD];
                  idx = (shift + pos - DIFGRD)%length;
                  idxp = (shift + pos + DIFGRD)%length;
                  v2 = p2[idx];
                  v2p = p2[idxp];
                  res2 += fabs((v1-v1p)-(v2-v2p)); }
             //res2/=length;

             if (resF==-1 || res2<resF)  resF=res2; }
        return resF; } 

   // LMoy Distance 
   //--------------------- 
    if (DistMap==DIST_LMoy)
       {
	float MoyV1=0,MoyV2=0;
  
        for(pos=0; pos<length; pos++) 
	   { MoyV1+=p1[pos];
             MoyV2+=p2[pos]; }

        resF=fabs(MoyV1-MoyV2);
        //resF/=length;
        return resF; }

   // TEST::[1]
   //--------------------- 
   if (DistMap==8)
      { for(shift=0; shift<SHIFT_DIST; shift+=STEP_SHIFT)
           { res2=0.0;
             for(pos=0; pos<(length-SHIFT_DIST); pos++)
                { v1 = p1[pos];
                  idx = (shift + pos)%length;
                  v2 = p2[idx];
                  res2 += (v1-v2)*(v1-v2); }
   	     //res2/=length; 

             if (resF==-1 || res2<resF)  resF=res2; }
        return sqrt(resF); }

   // TEST::[2]
   //--------------------- 
   if (DistMap==9)
      { for(shift=0; shift<30; shift+=2)
           { res2=0.0;
             for(pos=0; pos<(length-30); pos++)
                { v1 = p1[pos];
                  idx = (shift + pos)%length;
                  v2 = p2[idx];
                  res2 += (v1-v2)*(v1-v2); }
   	     //res2/=length; 

             if (resF==-1 || res2<resF)  resF=res2; }
        return sqrt(resF); }

 //Return
 return resF;
}


//-----------------------------
//-----------------------------
// INFO -----------------------
//-----------------------------
//-----------------------------
//------------------------------------------------
// FilInfoDist
//------------------------------------------------
void MdsBasedSymEst::FillInfoNameDist()
{
 strcpy(Info,Name);

 // if (DistMap==DIST_L2NS) strcat(Info,"_L2NS");
 // if (DistMap==DIST_L2)   strcat(Info,"_L2");
 // if (DistMap==DIST_L1)   strcat(Info,"_L1");
 // if (DistMap==DIST_LInf) strcat(Info,"_LInf");
 // if (DistMap==DIST_LMod) strcat(Info,"_LMod");
 // if (DistMap==DIST_LRad) strcat(Info,"_LRad");
 // if (DistMap==DIST_LGrd) strcat(Info,"_LCor");
 // if (DistMap==DIST_LMoy) strcat(Info,"_LMoy");

 if (DistMap==DIST_L2NS) strcat(Info,"_d0");
 if (DistMap==DIST_L2)   strcat(Info,"_d1");
 if (DistMap==DIST_L1)   strcat(Info,"_d2");
 if (DistMap==DIST_LInf) strcat(Info,"_d3");
 if (DistMap==DIST_LMod) strcat(Info,"_d4");
 if (DistMap==DIST_LRad) strcat(Info,"_d5");
 if (DistMap==DIST_LGrd) strcat(Info,"_d6");
 if (DistMap==DIST_LMoy) strcat(Info,"_d7");
}

//------------------------------------------------
// FilInfoDist
//------------------------------------------------
void MdsBasedSymEst::GiveResumeInfo()
{
 char BufTmp1[NCHAR];
 char BufTmp2[NCHAR];
 char BufTmp3[NCHAR];
 char BufTmp4[NCHAR];
 char BufTmp5[NCHAR];

 //Info
 printf("\n");
 printf("\n Resume Info sur [%s]",Name);
 printf("\n ---------------------------");
 printf("\n > Depth_[%d]:::Refining_[%d]:::Stochastique_[%d]",depth,FlgRefining,FlgStoch);
 printf("\n > Dist_[%d]:::FactPrior_[%.4f]",DistMap,FactPrior);
 printf("\n");
 printf("\n > ElapsedTime_[%.0f secondes]",ElapsedTime);
 printf("\n > PourcPixSatRGB_[%.3f]",PourcPixSatRGB);
 printf("\n");
 printf("\n > Corel_[%.3f > %.3f]",CorelRaw,CorelRefined);
 printf("\n > Nrj_[%.2f > %.3f]",NrjBeg,NrjEnd);
 printf("\n > ASI_[%.2f]",ASIndx);
 printf("\n");

 //Record_In_Name
 strcpy(Info,Name);
 sprintf(BufTmp1,"_Dpth_%d_Refin_%d_Stoc_%d_Prior_%d",depth,FlgRefining,FlgStoch,(int)(FactPrior*1000));
 strcat(Info,BufTmp1);
 sprintf(BufTmp2,"_Cor_%d_%d",int(CorelRaw*100),int(CorelRefined*100));
 strcat(Info,BufTmp2);
 sprintf(BufTmp3,"_Nrj_%d_%d",int(NrjBeg*1000),int(NrjEnd*1000));
 strcat(Info,BufTmp3);;
 sprintf(BufTmp4,"_ASI_%d_Dist_%d",int(ASIndx*100),DistMap);
 strcat(Info,BufTmp4);

 //InfoSup
 sprintf(BufTmp5,"_E%s_",Extension);
 strcat(Info,BufTmp5);
 
 //Name
 printf("\n Info File > [%s]",Info);
 printf("\n\n");
}

//------------------------------------------------
// SaveCurveASI
//------------------------------------------------
void MdsBasedSymEst::SaveCurveASI()
{
 int l;
 char BufferFile[NCHAR];

 //NameFile
 strcpy(BufferFile,"FileASI_");
 //strcat(BufferFile,Info);
 strcat(BufferFile,ExtensionASI);
 strcat(BufferFile,".dat");

 //Save ASI_curve
 FILE *fileASI=fopen(BufferFile,"w");
 //for(l=0;l<height;l++) fprintf(fileASI,"%d %d\n",l,(int) ASIcurve[l]); 
 //for(l=0;l<height;l++) fprintf(fileASI,"%d\n",(int) ASIcurve[l]);  
 for(l=PATIENT_YMIN;l<PATIENT_YMAX;l++) fprintf(fileASI,"%d\n",(int)ASIcurve[l]);     
 fclose(fileASI);
}

//----------------------------------------------------------
// Sauvegarde de l'image de nom <name> au format ppm       
//
// #define PATIENT_XMIN  155
// #define PATIENT_XMAX  380
// #define PATIENT_YMIN   28
// #define PATIENT_YMAX  620
// 
//----------------------------------------------------------
void  MdsBasedSymEst::SaveImagePpm(const char* Path_out,char* Name,char* Com,unsigned short*** matrvb,int lgth,int wdth)
{
    int i,j,k;
    int LgthImgCropped,WdthImgCropped;
    char buff[200];
    FILE* fuser;
    const int DEC_XMAX=-30;

    //Extension
    strcpy(buff,Path_out);
    strcat(buff,Name);
    strcat(buff,Com);
    strcat(buff,".ppm");

    //Ouverture fichier
    fuser=fopen(buff,"w");
    if (fuser==NULL)
    { printf(" probleme dans la sauvegarde de %s",buff);
        exit(-1); }

    //Affichage
    printf(" > Sauvegarde dans -> %s au format %s",buff,"ppm");
    printf(" [%d::%d]>[%d::%d]\n",lgth,wdth,LgthImgCropped,WdthImgCropped);
    fflush(stdout);

    //Long_&_Largeur
    LgthImgCropped=PATIENT_YMAX-PATIENT_YMIN;
    WdthImgCropped=PATIENT_XMAX-PATIENT_XMIN+DEC_XMAX;

    //Sauvegarde de l'entete
    fprintf(fuser,"P6");
    fprintf(fuser,"\n# %s \n",Com);
    fprintf(fuser,"%d %d",WdthImgCropped,LgthImgCropped);
    fprintf(fuser,"\n255\n");

    for(k=0;k<3;k++) 
        for(i=PATIENT_YMIN;i<PATIENT_YMIN+LgthImgCropped;i++) 
           for(j=PATIENT_XMIN;j<PATIENT_XMIN+WdthImgCropped;j++)
               { matrvb[k][i][j]=(int)(matrvb[k][i][j]+0.49);
                 if (matrvb[k][i][j]>255) matrvb[k][i][j]=255; }

    //Enregistrement
    for(i=PATIENT_YMIN;i<PATIENT_YMIN+LgthImgCropped;i++) 
        for(j=PATIENT_XMIN;j<PATIENT_XMIN+WdthImgCropped;j++)
           { fprintf(fuser,"%c",(char)matrvb[0][i][j]);
             fprintf(fuser,"%c",(char)matrvb[1][i][j]);
             fprintf(fuser,"%c",(char)matrvb[2][i][j]); }

    //Fermeture fichier
    fclose(fuser);
}


//-----------------------------
//-----------------------------
// VISU CLUSTERS --------------
//-----------------------------
//-----------------------------
//------------------------------------------------
// VisuThreeClusters
//------------------------------------------------
void MdsBasedSymEst::VisuThreeClusters()
{
 int i,j,k;
 int h,w;

 //MemoryAllocation
 float** matx=fmatrix_allocate_2d((3*17),2);
 float*** mattmp=fmatrix_allocate_3d(2,3,17);

 //Initialization
 for(k=0;k<2;k++) for(i=0;i<3;i++) for(j=0;j<17;j++) matx[i*17+j][k]=0.0;

 matx[0*17+0][0]=2422;     matx[1*17+0][0]=2160;     matx[2*17+0][0]=2396;
 matx[0*17+1][0]=2122;     matx[1*17+1][0]=2782;     matx[2*17+1][0]=2855;
 matx[0*17+2][0]=2839;     matx[1*17+2][0]=3069;     matx[2*17+2][0]=3372;
 matx[0*17+3][0]=2860;     matx[1*17+3][0]=3427;     matx[2*17+3][0]=3677;
 matx[0*17+4][0]=2064;     matx[1*17+4][0]=3289;     matx[2*17+4][0]=3281;
 matx[0*17+5][0]=2516;     matx[1*17+5][0]=2541;     matx[2*17+5][0]=4616;
 matx[0*17+6][0]=2493;     matx[1*17+6][0]=2869;     matx[2*17+6][0]=2254;
 matx[0*17+7][0]=2538;     matx[1*17+7][0]=2515;     matx[2*17+7][0]=2800;
 matx[0*17+8][0]=2904;     matx[1*17+8][0]=3037;     matx[2*17+8][0]=2598;
 matx[0*17+9][0]=2854;     matx[1*17+9][0]=2500;     matx[2*17+9][0]=2270;
 matx[0*17+10][0]=1995;    matx[1*17+10][0]=3513;    matx[2*17+10][0]=2009;
 matx[0*17+11][0]=2268;    matx[1*17+11][0]=2762;    matx[2*17+11][0]=3153;
 matx[0*17+12][0]=2695;    matx[1*17+12][0]=3041;    matx[2*17+12][0]=2399;
 matx[0*17+13][0]=2679;    matx[1*17+13][0]=3337;    matx[2*17+13][0]=3728;
 matx[0*17+14][0]=2500;    matx[1*17+14][0]=2326;    matx[2*17+14][0]=2643;
 matx[0*17+15][0]=2516;    matx[1*17+15][0]=2541;    matx[2*17+15][0]=4616;
 matx[0*17+16][0]=2699;    matx[1*17+16][0]=3232;    matx[2*17+16][0]=2939;

 //Presentation
 printf(" : Start FastMap...");
 fflush(stdout);
 FastMap* fmp;
 fmp=new FastMap;
 (*fmp).MyShiftDistChoice=100; 
 (*fmp).MyFastMap_(matx,(3*17),1,2,0);
 printf("done.");
 fflush(stdout);

 //MDS-SymMap Result
 for(k=0;k<(3*17);k++)
    { h=(int)(k/17);
      w=(int)(k%17);
      mattmp[0][h][w]=matx[k][0];
      mattmp[1][h][w]=matx[k][1];  }

 //VisuResult
 for(i=0;i<3;i++) 
    {  printf("\n");
      for(j=0;j<17;j++) printf("\n %.2f %.2f",mattmp[0][i][j],mattmp[1][i][j]); }
 printf("\n\n");

 //FreeMemory_&_Object
 if (matx)   free_fmatrix_2d(matx);
 if (mattmp) free_fmatrix_3d(mattmp,2);
 if (fmp)    delete(fmp);
}

//-----------------------------
//------------------------------------------------
// ReduceCurveASI_2Dpts
//------------------------------------------------
void MdsBasedSymEst::ReduceCurveASI_2Dpts()
{
 int i,j,k,l;
 int h,w;
 int tmp;
 int flag;
 char BufferFile[NCHAR];

 //MemoryAllocation
 float** matx=fmatrix_allocate_2d((3*17),PATIENT_YMAX-PATIENT_YMIN);
 float*** mattmp=fmatrix_allocate_3d(2,3,17);
 
 //Loop
 for(k=1;k<=(3*17);k++)
   {  //printf("\n[%d]",k);
      //fflush(stdout);
      sprintf(BufferFile,"FileASI_%d.dat",k);
      //printf("[%s]",BufferFile);
      //fflush(stdout);
      FILE* fileASI=fopen(BufferFile,"r");
      for(l=PATIENT_YMIN;l<PATIENT_YMAX;l++) 
	 { flag=fscanf(fileASI,"%d\n",&tmp);
           //printf("[%d]",tmp);
           //fflush(stdout);
           matx[k-1][l-PATIENT_YMIN]=tmp; }
      //for(l=PATIENT_YMIN;l<PATIENT_YMAX;l++) printf("[%.2f]", matx[k-1][l-PATIENT_YMIN]);
      fclose(fileASI); }

 //Presentation
 printf(" : Start FastMap...[CurveASI>2Dpts][dist:%d]",DistMap);
 fflush(stdout);
 FastMap* fmp;
 fmp=new FastMap;
 (*fmp).MyShiftDistChoice=DistMap; 
 (*fmp).MyFastMap_(matx,(3*17),PATIENT_YMAX-PATIENT_YMIN,2,0);
 printf("done.");
 fflush(stdout); 

  //MDS-SymMap Result
 //for(k=0;k<(3*17);k++) { printf("\n[%.2f::%.2f]",matx[k][0],matx[k][1]); }
 //fflush(stdout);
 for(k=0;k<(3*17);k++)
    { h=(int)(k/17);
      w=(int)(k%17);
      mattmp[0][h][w]=matx[k][0];
      mattmp[1][h][w]=matx[k][1];  }

  //VisuResult
 for(i=0;i<3;i++) 
    {  printf("\n");
      for(j=0;j<17;j++) printf("\n%.2f %.2f",mattmp[0][i][j],mattmp[1][i][j]); }

 //FreeMemory_&_Object
 if (matx)   free_fmatrix_2d(matx);
 if (mattmp) free_fmatrix_3d(mattmp,2);
 if (fmp)    delete(fmp);
}
