//---------------------------------------------------
// module  : f_PreProcDepthMapSeq.cc                       
// auteur  : Max Mignotte                            
// date    :                                              
// langage : C++                                      
// labo    :                                  
//---------------------------------------------------

//-----------------------------------------------
// FICHIERS INCLUS ------------------------------
//-----------------------------------------------
#include "PreProcDepthMapSeq.h"

//----------------------------------------------------------
// Constructeur de la classe PreProcDepthMapSeq
//----------------------------------------------------------
PreProcDepthMapSeq::PreProcDepthMapSeq(char* Nm,int lgth,int wdth,int dpth)           
{
 //RecordParameters
 strcpy(Name,Nm);
 depth=dpth; 
 height=lgth;
 width=wdth;

 //MemoryAllocation
 RawDepthMapSeq=usimatrix_allocate_3d(dpth,lgth,wdth);
 ProcDepthMapSeq=usimatrix_allocate_3d(dpth,lgth,wdth);
}

//----------------------------------------------------------
// Destructeur de la classe PreProcDepthMapSeq
//----------------------------------------------------------
PreProcDepthMapSeq::~PreProcDepthMapSeq()
{ 
  //Libere Memoire Matrice 
  if (RawDepthMapSeq)  free_usimatrix_3d(RawDepthMapSeq,depth);
  if (ProcDepthMapSeq) free_usimatrix_3d(ProcDepthMapSeq,depth);
}

//------------------------------------------------
// METHODES DE LA CLASSE PreProcDepthMapSeq
//------------------------------------------------
//------------------------------------------------
// Segmente_Silouhette
//------------------------------------------------
void PreProcDepthMapSeq::Segmente_Silouhette(unsigned short*** seq_in,unsigned short*** seq_out,\
float norm_factor,int filt_wsize)
{
 int x,y,d;
 float tmp;

 //Camera Intrinsic Parameters
 float focal=KINECT_FOCAL;
 float c_v = height/2.f;    

 //2D Cropping Positions
 int x_min=PATIENT_XMIN;
 int x_max=PATIENT_XMAX;
 int y_min=PATIENT_YMIN;
 int y_max=PATIENT_YMAX;

 float z_min=PATIENT_ZMIN;
 float z_max=PATIENT_ZMAX;
 float z_temp,z_mid;  

 //3D Vertical Treshold for the Treadmill
 float Ytresh=Y_TREADMILL_TRESH;
 float treadmill_y;  //3D vertical postition of the treadmill

 //Record Data Before Processing
 for(d=0;d<depth;d++) for(y=0;y<height;y++) for(x=0;x<width;x++) 
   { RawDepthMapSeq[d][y][x]=seq_in[d][y][x];
     ProcDepthMapSeq[d][y][x]=seq_in[d][y][x]; }


 //[1]-Estimate>z_mid
 //------------------------------------
 double mean=0.0;
 int counter=0;

 for(d=0;d<depth;d++) 
 for(y=0;y<height;y++) for(x=0;x<width;x++)
    {
     z_temp=ProcDepthMapSeq[d][y][x];
     treadmill_y=(focal*Ytresh+c_v*z_temp)/z_temp;

     if( y<y_min || x<x_min || x>x_max || y>y_max || z_temp<z_min || z_temp>z_max )
     continue;                          // Outside the 3D box

     else if (y>treadmill_y) continue;  // It's the treadmill
     else mean+=z_temp;                 // It's probably the patient
         
     counter++;
     }

    z_mid=mean/(counter);
    printf(" (Z_mid: [%.1f])\n",z_mid);


  //[2]-Clipping_Step
  //---------------------------------
  for(d=0;d<depth;++d)
  for(y=0;y<height;++y) for(x=0;x<width;++x)
     {
      z_temp=ProcDepthMapSeq[d][y][x];
      treadmill_y=(focal*Ytresh+c_v*z_temp)/z_temp;
                
      if( y<y_min || x<x_min || x>x_max || y>y_max || z_temp<z_min || z_temp>z_max )
      ProcDepthMapSeq[d][y][x]=(unsigned short)z_mid;                         // It's outside the patient 3D box
                
      else if (y>treadmill_y) ProcDepthMapSeq[d][y][x]=(unsigned short)z_mid; // It's the treadmill
     }

  //[3]-Normalization_&_Transfert
  //--------------------------------
  for(d=0;d<depth;++d)
  for(y=0;y<height;++y) for(x=0;x<width;++x)
     {
       tmp=ProcDepthMapSeq[d][y][x];
       tmp-=z_min;
       tmp*=((norm_factor/(z_max-z_min)));
       if (tmp>255) tmp=255;
       if (tmp<0)   tmp=0;

       ProcDepthMapSeq[d][y][x]=(unsigned char)tmp;
     }

  //[4]-Median_Filtering_Step
  //---------------------------------
  if (FlagMedian)
     { printf("\n > Start Median Filtering (filter size[%d]. Wait.) ...",filt_wsize);
       fflush(stdout);
       MedianFilter3D(ProcDepthMapSeq,depth,height,width,filt_wsize);
       printf(" done.\n"); }

  //Transfert
  for(d=0;d<depth;++d)
  for(y=0;y<height;++y) for(x=0;x<width;++x) seq_out[d][y][x]=ProcDepthMapSeq[d][y][x];
}
